import ast
import logging
import re
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
from egse.setup import navdict

logger = logging.getLogger(__name__)


def dict_to_ref_model(model_def: Union[Dict, List]) -> navdict:
    """Creates a reference frames model from a dictionary or list of reference frame definitions.

    When a list is provided, the items in the list must be ReferenceFrames.

    The reference frame definitions are usually read from a YAML file or returned by a Setup, but can also be just
    ReferenceFrame objects.

    ReferenceFrame definitions have the following format:

    ```
    ReferenceFrame://(<definition>)
    ```
    where `<definition>` has the following elements, separated by '` | `':
    * a translation matrix
    * a rotation matrix
    * the name of the reference frame
    * the name of the reference for this reference frame
    * a dictionary of links

    Args:
        model_def (dict or list): Definition of the reference model.

    Returns:
        Dictionary representing the reference frames model.
    """

    ref_model = navdict()
    ref_links = {}

    from egse.coordinates.reference_frame import ReferenceFrame

    def create_ref_frame(name, data) -> Union[ReferenceFrame, str]:
        # This is a recursive function that creates a reference frame based on the given data.
        # * When the data is already a ReferenceFrame, it just returns data
        # * When data starts with the special string `ReferenceFrame//`, the data string is parsed
        #   and a corresponding ReferenceFrame is returned
        # * When there is no match, the data is returned unaltered.
        #
        # SIDE EFFECT:
        # * In the process, the outer ref-model and ref_links are updated.

        if isinstance(data, ReferenceFrame):
            return data

        match = re.match(r"ReferenceFrame//\((.*)\)$", data)
        if not match:
            return data

        translation, rotation, name, ref_name, links = match[1].split(" | ")

        # All links are processed later

        ref_links[name] = ast.literal_eval(links)

        if ref_name == name == "Master":
            ref_model.add(ref_name, ReferenceFrame.create_master())
            return ref_model["Master"]

        if ref_name not in ref_model:
            ref_model.add(ref_name, create_ref_frame(ref_name, model_def[ref_name]))

        ref_frame = ReferenceFrame.from_translation_rotation(
            deserialize_array(translation),
            deserialize_array(rotation),
            name=name,
            ref=ref_model[ref_name],
        )

        return ref_frame

    # if the given model_def is a list, turn it into a dict

    if isinstance(model_def, list):
        model_def = {frame.name: frame for frame in model_def}

    for key, value in model_def.items():
        if key not in ref_model:
            ref_model.add(key, create_ref_frame(key, value))

    # Process all the links

    for ref_name, link_names in ref_links.items():
        ref = ref_model[ref_name]
        for link_name in link_names:
            if link_name not in ref.linked_to:
                ref.add_link(ref_model[link_name])

    return ref_model


def ref_model_to_dict(ref_model) -> navdict:
    """Creates a dictionary with reference frames definitions that define a reference model.

    Args:
        ref_model: A dictionary representing the reference frames model or a list of reference frames.

    Returns:
        Dictionary of reference frame definitions.
    """

    if isinstance(ref_model, dict):
        ref_model = ref_model.values()

    # take each key (which is a reference frame) and serialize it

    model_def = {}

    for ref in ref_model:
        translation, rotation = ref.get_translation_rotation_vectors()
        links = [ref.name for ref in ref.linked_to]
        model_def[ref.name] = (
            f"ReferenceFrame//("
            f"{serialize_array(translation, precision=6)} | "
            f"{serialize_array(rotation, precision=6)} | "
            f"{ref.name} | "
            f"{ref.reference_frame.name} | "
            f"{links})"
        )

    return navdict(model_def)


def serialize_array(arr: Union[np.ndarray, list], precision: int = 4) -> str:
    """Returns a string representation of a numpy array.

    >>> serialize_array([1,2,3])
    '[1, 2, 3]'
    >>> serialize_array([[1,2,3], [4,5,6]])
    '[[1, 2, 3], [4, 5, 6]]'
    >>> serialize_array([[1,2.2,3], [4.3,5,6]])
    '[[1.0000, 2.2000, 3.0000], [4.3000, 5.0000, 6.0000]]'
    >>> serialize_array([[1,2.2,3], [4.3,5,6]], precision=2)
    '[[1.00, 2.20, 3.00], [4.30, 5.00, 6.00]]'

    Args:
        arr: One- or-two dimensional numpy array or list.
        precision (int): number of digits of precision
    Returns:
        A string representing the input array.
    """
    if isinstance(arr, list):
        arr = np.array(arr)
    msg = np.array2string(
        arr,
        separator=", ",
        suppress_small=True,
        formatter={"float_kind": lambda x: f"{x:.{precision}f}"},
    ).replace("\n", "")
    return msg


def deserialize_array(arr_str: str) -> Optional[np.ndarray]:
    """Returns a numpy array from the given string.

    The input string is interpreted as a one or two-dimensional array, with commas or spaces separating the columns,
    and semicolons separating the rows.

    >>> deserialize_array('1,2,3')
    array([1, 2, 3])
    >>> deserialize_array('1 2 3')
    array([1, 2, 3])
    >>> deserialize_array('1,2,3;4,5,6')
    array([[1, 2, 3],
           [4, 5, 6]])
    >>> deserialize_array("[[1,2,3], [4,5,6]]")
    array([[1, 2, 3],
           [4, 5, 6]])

    Args:
        arr_str: String representation of a numpy array.

    Returns:
        One- or two-dimensional numpy array or `None` when input string cannot be parsed into a numpy array.
    """

    import re

    arr_str = re.sub(r"\],\s*\[", "];[", arr_str)
    try:
        arr = np.array(_convert_from_string(arr_str))
        return arr if ";" in arr_str else arr.flatten()
    except ValueError as exc:
        logger.error(f"Input string could not be parsed into a numpy array: {exc}")
    return None


def _convert_from_string(data: str) -> list[list]:
    # This function was copied from:
    #   https://github.com/numpy/numpy/blob/v1.19.0/numpy/matrixlib/defmatrix.py#L14
    # We include the function here because the np.matrix class is deprecated and will be removed.
    # This function is what we actually needed from np.matrix.

    # This function can be replaced with np.fromstring()

    for char in "[]":
        data = data.replace(char, "")

    rows = data.split(";")
    new_data = []
    count = 0
    for row in rows:
        trow = row.split(",")
        new_row = []
        for col in trow:
            temp = col.split()
            new_row.extend(map(ast.literal_eval, temp))
        if count == 0:
            n_cols = len(new_row)
        elif len(new_row) != n_cols:
            raise ValueError("Rows not the same size.")
        count += 1
        new_data.append(new_row)

    return new_data



import datetime
import os
import textwrap
from pathlib import Path

import rich

from egse.env import get_site_id
from egse.hk import get_housekeeping
from egse.setup import Setup
from egse.system import EPOCH_1958_1970
from egse.system import format_datetime
from fixtures.helpers import create_text_file

HERE = Path(__file__).parent


def test_get_housekeeping(default_env):
    data_dir = Path(default_env.data_root)

    # The get_housekeeping() function will by default search for HK telemetry in the daily CSV file that contains the
    # HK parameter that is passed into the function. Since we have no operational system running during the tess,
    # we will need to create sample data for that parameter in the correct file. The file will be located in the
    # data storage location folder (defined by the default_env fixture) at
    # `HERE/data/LAB23/daily/YYYYMMDD/YYYYMMDD_LAB23_DAQ-TM.csv`, where 'LAB23' is the SITE_ID and 'DAQ-TM' is the
    # storage mnemonic of the device (this is read from the TM dictionary file).

    today = format_datetime("today")
    today_with_dash = format_datetime("today", fmt="%Y-%m-%d")

    tm_dictionary_path = data_dir / "../common/telemetry/tm-dictionary.csv"

    create_text_file(
        tm_dictionary_path,
        "TM source;Storage mnemonic;CAM EGSE mnemonic;Original name in EGSE;Name of corresponding timestamp;"
        "Origin of synoptics at CSL;Origin of synoptics at CSL1;Origin of synoptics at CSL2;"
        "Origin of synoptics at SRON;Origin of synoptics at IAS;Origin of synoptics at INTA;"
        "Description;MON screen;unit cal1;offset b cal1;slope a cal1;calibration function;"
        "MAX nonops;MIN nonops;MAX ops;MIN ops;Comment\n"
        "Unit Test Manager;DAQ-TM;TEMP_ABC_001;ABC_001;timestamp;;;;;;;Temperature of ABC;;;;;;;;;;\n",
        create_folder=True,
    )

    hk_path = data_dir / f"daily/{today}/{today}_{get_site_id()}_DAQ-TM.csv"

    create_text_file(
        hk_path,
        textwrap.dedent(
            f"""\
            timestamp,TEMPT_ABC_000,TEMP_ABC_001,TEMP_ABC_002
            {today_with_dash}T00:00:23.324+0000,21.333,23.3421,26.234
            {today_with_dash}T00:00:42.123+0000,22.145,23.4567,27.333
            """
        ),
        create_folder=True,
    )

    # The TM dictionary will be loaded relative from the location of the Setup YAML file.
    # Since we read the Setup from a string, there is no location and the dictionary will
    # be loaded from the default resource location or from the current working directory '.'.

    os.environ["NAVDICT_DEFAULT_RESOURCE_LOCATION"] = str(data_dir)

    setup = Setup.from_yaml_string(
        textwrap.dedent(
            """
            telemetry:
                dictionary: pandas//../common/telemetry/tm-dictionary.csv
                dictionary_kwargs:
                    separator: ;
            """
        )
    )

    rich.print(f"{hk_path = }")
    rich.print(f"{tm_dictionary_path = }")

    try:
        timestamp, data = get_housekeeping("TEMP_ABC_001", setup=setup)
        timestamp -= EPOCH_1958_1970
        dt = datetime.datetime.utcfromtimestamp(timestamp)

        rich.print(f"{timestamp}, {dt}, {data}")

        assert data.strip() == "23.4567"
        assert format_datetime(dt, fmt="%Y-%m-%d").startswith(today_with_dash)

    finally:
        ...
        # Get rid of the CSV file
        # hk_path.unlink()

        # Get rid of the tm dictionary file
        # tm_dictionary_path.unlink()


def test_convert_hk_names():
    a = {
        "aaa": 1,
        "bbb": 2,
        "ccc": 3,
        "eee": 4,
    }

    c = {
        "aaa": "AAA",
        "bbb": "BBB",
        "ccc": "CCC",
        "ddd": "DDD",
    }

    from egse.hk import convert_hk_names

    b = convert_hk_names(a, c)

    # Result:
    #  * all keys in 'a' that have a conversion in 'c' shall be in 'b' with the converted key
    #  * all keys in 'a' that do not have a conversion in 'c', shall be in 'b' with their original key
    #  * all conversion keys that are in 'c' but not in 'a' shall just be ignored

    assert "AAA" in b
    assert "BBB" in b
    assert "CCC" in b
    assert "eee" in b

    assert "aaa" not in b
    assert "bbb" not in b
    assert "ccc" not in b
    assert "ddd" not in b
    assert "DDD" not in b

    for k, v in a.items():
        if k == "eee":
            assert b[k] == v
        else:
            assert b[k.upper()] == v



import logging
from pathlib import Path
from typing import Dict, List, Tuple

from egse.connect import get_endpoint
from egse.decorators import dynamic_interface
from egse.device import DeviceConnectionState, DeviceInterface
from egse.env import bool_env
from egse.response import Failure
from egse.log import logger
from egse.mixin import CommandType, DynamicCommandMixin, add_lf, dynamic_command
from egse.proxy import DynamicProxy
from egse.scpi import count_number_of_channels, create_channel_list, get_channel_names
from egse.settings import Settings
from egse.tempcontrol.keithley.daq6510_dev import DAQ6510

VERBOSE_DEBUG = bool_env("VERBOSE_DEBUG")

HERE = Path(__file__).parent

cs_settings = Settings.load("Keithley Control Server")
dev_settings = Settings.load("Keithley DAQ6510")

PROTOCOL = cs_settings.get("PROTOCOL", "tcp")
HOSTNAME = cs_settings.get("HOSTNAME", "localhost")
COMMANDING_PORT = cs_settings.get("COMMANDING_PORT", 0)
TIMEOUT = cs_settings.get("TIMEOUT", 60)
SERVICE_TYPE = cs_settings.get("SERVICE_TYPE", "daq6510")

DEFAULT_BUFFER_1 = "defbuffer1"
DEFAULT_BUFFER_2 = "defbuffer2"

DEV_HOST = dev_settings.get("HOSTNAME", "localhost")
DEV_PORT = dev_settings.get("PORT", 5025)


def decode_response(response: bytes) -> str | Failure:
    """Decodes the bytes object, strips off the trailing 'CRLF'."""

    if VERBOSE_DEBUG:
        logger.debug(f"{response = } <- decode_response")

    if isinstance(response, Failure):
        return response

    if isinstance(response, memoryview):
        response = response.tobytes()

    return response.decode().rstrip()


class DAQ6510Interface(DeviceInterface):
    """
    Interface definition for the Keithley DAQ6510 Controller, Proxy, and Simulator.
    """

    @dynamic_interface
    def send_command(self, command: str, response: bool) -> str | None:
        """Sends the given SCPI command to the device.

        The valid commands are described in the DAQ6510 Reference Manual [DAQ6510-901-01 Rev. B / September 2019].

        Args:
            command (str): SCPI command as specified in the DAQ6510 Reference Manual
            response (bool): Indicates whether you expect a reply from the device

        Returns: Response from the DAQ6510 is returned when a response was expected. When `response` is False, None
                 will be returned.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="*IDN?",
        process_cmd_string=add_lf,
        process_response=decode_response,
    )
    def get_idn(self) -> str:
        """Returns basic information about the device, its name, firmware version, etc.

        The string returned is subject to change without notice and can not be used for parsing information.

        Returns: Identification string of the instrument.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="*IDN?",
        process_cmd_string=add_lf,
        process_response=decode_response,
    )
    def info(self) -> str:
        """Returns basic information about the device, its name, firmware version, etc.

        The string returned is subject to change without notice and can not be used for parsing information.

        Returns: Identification string of the instrument.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="*RST",
        process_cmd_string=add_lf,
    )
    def reset(self) -> None:
        """Resets the DAQ6510.

        This returns the instrument to default settings, and cancels all pending commands.

        Note:
            The `reset()` method also deletes all the user-defined buffers. The two default buffers are cleared.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE, cmd_string=":SYST:TIME ${year}, ${month}, ${day}, ${hour}, ${minute}, ${second}"
    )
    def set_time(self, year: int, month: int, day: int, hour: int, minute: int, second: int) -> None:
        """Sets the absolute date and time for the device.

        Args:
            year (int): Year
            month (int): Month
            day (int): Day
            hour (int): Hour
            minute (int): Minute
            second (int): Second
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string=":SYST:TIME? 1", process_response=decode_response)
    def get_time(self) -> str:
        """Gets the date and time from the device in UTC.

        The returned string is of the format:

            <weekday> <month> <day> <hour>:<minute>:<second> <year>

        Returns: Data and time from the device in UTC.
        """

        raise NotImplementedError

    @dynamic_interface
    def read_buffer(self, start: int, end: int, buffer_name: str, elements: List[str]):
        """Reads specific data elements (measurements) from the given buffer.

        Args:
            start: (int) First index of the buffer that should be returned (>= 1)
            end (int): Last index of the buffer that should be returned
            buffer_name (str): Name of the buffer to read out
            elements (List[str]): List of elements from the buffer to include in the response

        Returns: List of all the readings.
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string='TRAC:ACTUAL? "${buffer_name}"')
    def get_buffer_count(self, buffer_name: str = DEFAULT_BUFFER_1):
        """Returns the number of data points in the specified buffer.

        Args:
            buffer_name (str): Name of the buffer
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string='TRACE:POINTS? "${buffer_name}"',
        process_cmd_string=add_lf,
    )
    def get_buffer_capacity(self, buffer_name: str = DEFAULT_BUFFER_1):
        """Returns the capacity of the specified buffer.

        Args:
            buffer_name (str): Name of the buffer
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.WRITE, cmd_string='TRACE:DELETE "${buffer_name}"')
    def delete_buffer(self, buffer_name: str) -> None:
        """Deletes the specified buffer.

        Args:
            buffer_name (str): Name of the buffer
        """

        raise NotImplementedError

    @dynamic_interface
    def clear_buffer(self, buffer_name: str) -> None:
        """Clears the given buffer.

        Args:
            buffer_name (str): Name of the buffer
        """

        raise NotImplementedError

    @dynamic_interface
    def create_buffer(self, buffer_name: str, size: int) -> None:
        """Creates a reading buffer with the given name and size.

        Args:
            buffer_name (str): Name of the buffer
            size (size): Maximum number of readings (size >= 10) [default: 1000]
        """

        raise NotImplementedError

    @dynamic_interface
    def configure_sensors(self, channel_list: str, *, sense: Dict[str, List[Tuple]]):
        """Configures the DAQ6510 to sense the specified channels.

        Args:
            channel_list (str): List of channels, as understood by the device
            sense (Dict[str], List[Tuple]): Dictionary with all the information on the configuration of the channels
        """

        raise NotImplementedError

    @dynamic_interface
    def setup_measurements(self, *, buffer_name: str = DEFAULT_BUFFER_1, channel_list: str):
        """Sets up the measurements for the given channel list.

        Args:
            buffer_name (str): Name of the buffer to use [default: defbuffer1]
            channel_list (str): Channels to read out
        """

        raise NotImplementedError

    @dynamic_interface
    def perform_measurement(
        self, *, buffer_name: str = DEFAULT_BUFFER_1, channel_list: str, count: int, interval: int
    ) -> list:
        """Performs the actual measurements.

        Args:
            buffer_name (str): Name of the buffer [default: defbuffer1]
            channel_list (str): List of channels, as understood by the device
            count (int): Number of measurements to perform
            interval (int): Interval between measurements [s]
        """

        raise NotImplementedError


class DAQ6510Controller(DAQ6510Interface, DynamicCommandMixin):
    """
    The DAQ6510 Controller allows to remotely control the Keithley Data Acquisition System
    through an Ethernet interface.
    """

    def __init__(self, hostname: str = DEV_HOST, port: int = DEV_PORT):
        """Opens a TCP/IP socket connection with the Keithley DAQ6510 Hardware.

        Args:
            hostname (str): IP address or fully qualified hostname of the DAQ6510 hardware controller.
                The default is defined in the ``settings.yaml`` configuration file.
            port (int): IP port number to connect to, by default set in the ``settings.yaml`` configuration file.

        Raises:
            Error: when the connection could not be established for some reason.
        """

        super().__init__()

        logger.debug(f"Initializing the DAQ6510 Controller with hostname={hostname} on port={port}")

        self.daq = self.transport = DAQ6510(hostname, port)

        # We set the default buffer here, this can be changed with the `create_buffer()` method.

        self.buffer_name = DEFAULT_BUFFER_1

    def is_simulator(self) -> bool:
        """Indicates that the device is a real hardware controller

        Returns: False.
        """

        return False

    def connect(self) -> None:
        """Connects to the device controller."""

        self.daq.connect()
        self.notify_observers(DeviceConnectionState.DEVICE_CONNECTED)

    def disconnect(self) -> None:
        """Disconnects from the device controller."""

        self.daq.disconnect()
        self.notify_observers(DeviceConnectionState.DEVICE_NOT_CONNECTED)

    def reconnect(self) -> None:
        """Reconnects to the device controller."""

        if self.is_connected():
            self.disconnect()

        self.connect()

    def is_connected(self) -> bool:
        """Checks whether the device controller is connected.

        Returns: True if the device controller is connected; False otherwise.
        """

        return self.daq.is_connected()

    def send_command(self, command: str, response: bool) -> str | None:
        """Sends an SCPI command to the device.

        The valid commands are described in the DAQ6510 Reference Manual [DAQ6510-901-01 Rev. B / September 2019].

        Args:
            command (str): SCPI command as specified in the DAQ6510 Reference Manual
            response (bool): Indicates whether you expect a reply from the device

        Returns: The response from the DAQ6510 is returned when a response was expected. When `response` is False,
                 None will be returned.
        """

        return self.daq.trans(command).decode() if response else self.daq.write(command)

    def read_buffer(
        self, start: int, end: int, buffer_name: str = DEFAULT_BUFFER_1, elements: list[str] | None = None
    ) -> bytes:
        """Reads specific data elements (measurements) from the given buffer.

        Elements that can be specified to read out:

            - CHANNEL: Channel for which the data was acquired
            - DATE: Date when the data point was measured
            - READING: Actual reading of the measurement
            - TSTAMP: Timestamp when the data point was measured
            - UNIT: Unit of measure for the measurement
            - STATUS: Status information associated with the measurement

        Args:
            start: (int) First index of the buffer that should be returned (>= 1)
            end (int): Last index of the buffer that should be returned
            buffer_name (str): Name of the buffer to read out
            elements (list[str] | None): List of elements from the buffer to include in the response

        Returns: List of all the readings.
        """

        if elements is None:
            elements_str = ["READING"]
        else:
            elements_str = ", ".join(elements)

        return self.daq.trans(f'TRACE:DATA? {start}, {end}, "{buffer_name}", {elements_str}')

    def clear_buffer(self, buffer_name: str = DEFAULT_BUFFER_1) -> None:
        """Clears the given buffer.

        Args:
            buffer_name (str): Name of the buffer
        """

        response = self.daq.trans(f'TRACE:ACTUAL? "{buffer_name}"')

        logger.info(f"Clearing buffer '{buffer_name}' containing {response} readings.")

        self.daq.write(f'TRACE:CLEAR "{buffer_name}"')

    def create_buffer(self, buffer_name: str, size: int = 1000) -> None:
        """Creates a reading buffer with the given name.

        The name of the buffer must adhere to the following rules:

            - A buffer with this name should not exist in the device yet. When the buffer does exist, the DAQ6510 will
              show  a dialogue on the front panel with error 1115 saying the command cannot take an existing buffer
              name.
            - Buffer names must start with an alphabetic character.
            - The names cannot contain any periods nor the underscore (_) character.
            - The name can be up to 31 characters long.

        If the given size is 0, the instrument creates the largest reading buffer possible based on the available
        memory when the buffer is created.

        Args:
            buffer_name (str): Name of the buffer
            size (size): Maximum number of readings (size >= 10) [default: 1000]
        """

        self.daq.write(f'TRACE:MAKE "{buffer_name}", {size}')

        self.buffer_name = buffer_name

    # def reset(self) -> None:
    #
    #     self.daq.write("SYSTem:BEEPer 500, 0.1; :*RST; :SYSTem:BEEPer 1000, 0.1\n")

    def configure_sensors(self, channel_list: str, *, sense: Dict[str, List[Tuple]]) -> None:
        """Configures the different sensors in the `channel_list`.

        Each sensor in the list will be configured according to the settings given in the `sense` dictionary.

        The following code will configure channels 101 and 102 as 4-wire transducers of type PT100.

        ```
        channel_list = create_channel_list(101, 102)

        sense = {
            "TEMPERATURE": [
                ("TRANSDUCER", "FRTD"),
                ("RTD:FOUR", "PT100"),
            ]
        }

        daq.configure_sensors(channel_list, sense=sense)
        ```

        The `sense` argument is a dictionary where the keys are function names like "TEMPERATURE" or "VOLTAGE:DC",
        and the values are a list of settings for that function. The list of settings is a list of tuples with the
        command and the value, e.g. "TRANSDUCER" is the settings command and "FRTD" is its value. The list of settings
        will be sent to the device in the order that they take in the list.

        Args:
            channel_list (str): Channels to configure
            sense: Dictionary with all the information on the configuration
        """

        if "TEMPERATURE" in sense:
            # Allowed settings for TEMPERATURE:
            #
            # - TEMPERATURE:APERTURE                  (@<channelList>)
            # - TEMPERATURE:AVERAGE:COUNT             (@<channelList>)
            # - TEMPERATURE:AVERAGE:STATE             (@<channelList>)
            # - TEMPERATURE:AVERAGE:TCONTROL          (@<channelList>)
            # - TEMPERATURE:AVERAGE:WINDOW            (@<channelList>)
            # - TEMPERATURE:AVERAGE:AZERO:STATE
            # - TEMPERATURE:DELAY:AUTO
            # - TEMPERATURE:DELAY:USER<N>
            # - TEMPERATURE:LINE:SYNC
            # - TEMPERATURE:NPLCYCLES
            # - TEMPERATURE:OCOMPENSATED
            # - TEMPERATURE:ODETECTOR
            # - TEMPERATURE:RELATIVE
            # - TEMPERATURE:RELATIVE:ACQUIRE
            # - TEMPERATURE:RELATIVE:STATE
            # - TEMPERATURE:RTD:ALPHA, BETA, DELTA      (@<channelList>)
            # - TEMPERATURE:RTD:ZERO, TWO, THREE, FOUR  (@<channelList>)
            # - TEMPERATURE:TCOUPLE:RJUNCTION:SIMULATED
            # - TEMPERATURE:TCOUPLE:RJUNCTION:RSELECT
            # - TEMPERATURE:TCOUPLE:TYPE
            # - TEMPERATURE:THERMISTOR      (@<channelList>)
            # - TEMPERATURE:TRANSDUCER      (@<channelList>)
            # - TEMPERATURE:UNIT            (@<channelList>)
            #

            # set the function to temperature

            self.daq.write(f'SENSE:FUNCTION "TEMPERATURE", {channel_list}')

            for cmd, value in sense["TEMPERATURE"]:
                self.daq.write(f"SENSE:TEMPERATURE:{cmd} {value}, {channel_list}")

    def setup_measurements(self, *, buffer_name: str = DEFAULT_BUFFER_1, channel_list: str) -> None:
        """Sets up the measurements for the given channel list.

        Args:
            buffer_name (str): Name of the buffer to use [default: defbuffer1]
            channel_list (str): Channels to read out
        """

        self.daq.write(f'ROUTE:SCAN:BUFFER "{buffer_name}"')
        self.daq.write(f"ROUTE:SCAN:CREATE {channel_list}")
        self.daq.write(f"ROUTE:CHANNEL:OPEN {channel_list}")
        _ = self.daq.trans(f"ROUTE:CHANNEL:STATE? {channel_list}")
        self.daq.write("ROUTE:SCAN:START:STIMULUS NONE")

    def perform_measurement(
        self,
        *,
        buffer_name: str = DEFAULT_BUFFER_1,
        channel_list: str,
        count: int = 1,
        interval: int = 2,
    ) -> list:
        """Performs the actual measurements.

        This function will wait until all measurements have completed, so be careful with the arguments `count` and
        `interval` as they will multiply into the number of seconds that you will have to wait for the response.

        Args:
            buffer_name (str): Name of the buffer
            channel_list (str): List of channels, as understood by the device
            count (int): Number of measurements to perform
            interval (int): Interval between measurements [s]

        Returns: List of readings.
        """

        # Set the number of times a scan is repeated

        self.daq.write(f"ROUTE:SCAN:COUNT:SCAN {count}")
        self.daq.write(f"ROUTE:SCAN:INTERVAL {interval}")  # [seconds]

        #
        self.daq.write("INITIATE:IMMEDIATE")
        self.daq.write("*WAI")

        # Read out the buffer

        logger.debug(f"Buffer count = {self.get_buffer_count()}")

        num_sensors = count_number_of_channels(channel_list)

        readings = []

        for idx in range(1, count * num_sensors + 1):
            response = self.read_buffer(
                idx, idx, buffer_name=buffer_name, elements=["CHANNEL", "TSTAMP", "READING", "UNIT"]
            )
            response_str = response.decode().strip()

            if response_str != "" and response_str != str(count * num_sensors):
                if "\n" in response_str:
                    response_str = response_str.split("\n")
                    for i in range(len(response_str)):
                        readings.append(response_str[i].split(","))
                else:
                    readings.append(response_str.split(","))

                # Remove incomplete readings
                if len(readings[0]) < 4:
                    del readings[0]

        return readings


class DAQ6510Proxy(DynamicProxy, DAQ6510Interface):
    """
    The DAQ6510Proxy class is used to connect to the Keithley Control Server and send commands
    to the Keithley Hardware Controller remotely.
    """

    def __init__(
        self,
        protocol: str = PROTOCOL,
        hostname: str = HOSTNAME,
        port: int = COMMANDING_PORT,
        timeout: float = TIMEOUT,  # Timeout [s]: > scan count * interval + (one scan duration)
    ):
        """Initialization of a DAQ6510Proxy.

        Args:
            protocol (str): Transport protocol [default is taken from settings file]
            hostname (str): Location of the Control Server (IP address) [default is taken from settings file]
            port (int): TCP port on which the Control Server is listening for commands [default is taken from settings
                        file]
            timeout (float): Timeout by which to establish the connection [s]
        """

        endpoint = get_endpoint(SERVICE_TYPE, protocol, hostname, port)

        super().__init__(endpoint, timeout=timeout)



import asyncio
import datetime
import json
import signal
import time
from asyncio import Task
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Optional

import typer
import zmq
import zmq.asyncio

from egse.device import DeviceConnectionError
from egse.device import DeviceTimeoutError
from egse.log import logger
from egse.settings import Settings
from egse.system import TyperAsyncCommand
from egse.tempcontrol.keithley.daq6510_adev import DAQ6510

settings = Settings.load("Keithley DAQ6510")

DAQ_DEV_HOST = settings.get("HOSTNAME", "localhost")
DAQ_DEV_PORT = settings.get("PORT", 5025)

DAQ_MON_CMD_PORT = 5556


class DAQ6510Monitor:
    """
    DAQ6510 temperature monitoring service with ZeroMQ command interface.

    """

    def __init__(
        self,
        daq_hostname: str,
        daq_port: int = DAQ_DEV_PORT,
        zmq_port: int = DAQ_MON_CMD_PORT,
        log_file: str = "temperature_readings.log",
        channels: list[str] | None = None,
        poll_interval: float = 60.0,
    ):
        """Initialize the DAQ6510 monitoring service.

        Args:
            daq_hostname: Hostname or IP of the DAQ6510
            daq_port: TCP port for DAQ6510 SCPI interface
            zmq_port: Port for ZeroMQ command interface
            log_file: Path to log file for temperature readings
            channels: List of channels to monitor (e.g. ["101", "102"])
            poll_interval: Initial polling interval in seconds
        """
        self.daq_hostname = daq_hostname
        self.daq_port = daq_port
        self.zmq_port = zmq_port
        self.log_file = Path(log_file)
        self.channels = channels or ["101", "102", "103", "104"]
        self.poll_interval = poll_interval

        # Setup ZeroMQ context
        self.ctx = zmq.asyncio.Context()
        self.socket = self.ctx.socket(zmq.ROUTER)
        self.socket.bind(f"tcp://*:{zmq_port}")

        # Service state
        self.running = False
        self.polling_active = False
        self.daq_interface = None
        self.command_handlers: dict[str, Callable] = {
            "START_POLLING": self._handle_start_polling,
            "STOP_POLLING": self._handle_stop_polling,
            "SET_INTERVAL": self._handle_set_interval,
            "SET_CHANNELS": self._handle_set_channels,
            "GET_STATUS": self._handle_get_status,
            "GET_READING": self._handle_get_reading,
            "GET_LAST_READING": self._handle_get_last_reading,
            "SHUTDOWN": self._handle_shutdown,
        }

        # Keep a record of the last measurement
        self._last_reading: dict = {}

        # Make sure the log directory exists
        self.log_file.parent.mkdir(exist_ok=True, parents=True)

        # Create DAQ interface
        # In this case we use the device itself, no control server. That means
        # the monitoring must be the only service connecting to the device.
        self.daq_interface = DAQ6510(hostname=daq_hostname, port=daq_port)

    async def start(self):
        """Start the monitoring service."""
        logger.info(f"Starting DAQ6510 Monitoring Service on ZMQ port {self.zmq_port}")
        self.running = True

        def handle_shutdown():
            asyncio.create_task(self.shutdown())

        # Register signal handlers for graceful shutdown
        for sig in (signal.SIGINT, signal.SIGTERM):
            asyncio.get_event_loop().add_signal_handler(sig, handle_shutdown)

        # Start the main service tasks
        await asyncio.gather(self.command_listener(), self.connect_daq(), return_exceptions=True)

    def done_polling(self, task: Task):
        if task.exception():
            logger.error(f"Polling loop ended unexpectedly: {task.exception()}")
        logger.info(f"Done polling ({task.get_name()}).")
        self.polling_active = False

    async def connect_daq(self):
        """Establish connection to the DAQ6510."""
        while self.running:
            assert self.daq_interface is not None  # extra check and make mypy happy

            init_commands = [
                ('TRAC:MAKE "test1", 1000', False),  # create a new buffer
                # settings for channel 1 and 2 of slot 1
                ('SENS:FUNC "TEMP", (@101:102)', False),  # set the function to temperature
                ("SENS:TEMP:TRAN FRTD, (@101)", False),  # set the transducer to 4-wire RTD
                ("SENS:TEMP:RTD:FOUR PT100, (@101)", False),  # set the type of the 4-wire RTD
                ("SENS:TEMP:TRAN RTD, (@102)", False),  # set the transducer to 2-wire RTD
                ("SENS:TEMP:RTD:TWO PT100, (@102)", False),  # set the type of the 2-wire RTD
                ('ROUT:SCAN:BUFF "test1"', False),
                ("ROUT:SCAN:CRE (@101:102)", False),
                ("ROUT:CHAN:OPEN (@101:102)", False),
                ("ROUT:STAT? (@101:102)", True),
                ("ROUT:SCAN:STAR:STIM NONE", False),
                # ("ROUT:SCAN:ADD:SING (@101, 102)", False),  # not sure what this does, not really needed
                ("ROUT:SCAN:COUN:SCAN 1", False),  # not sure if this is needed in this setting
                # ("ROUT:SCAN:INT 1", False),
            ]

            try:
                logger.info(f"Connecting to DAQ6510 at {self.daq_hostname}:{self.daq_port}")
                await self.daq_interface.connect()
                logger.info("Successfully connected to DAQ6510.")
                await self.daq_interface.initialize(commands=init_commands, reset_device=True)
                logger.info("Successfully initialized DAQ6510 for measurements.")

                # If we were polling before, restart it.
                # The first time we enter this loop, we are not polling.
                if self.polling_active:
                    # QUESTION: Do we need to await here?
                    polling_task = asyncio.create_task(self.polling_loop())

                    # But we can add error handling for the task
                    polling_task.add_done_callback(self.done_polling)

                # Keep checking connection status periodically
                while self.running and await self.daq_interface.is_connected():
                    logger.info("Checking DAQ6510 connection...")
                    await asyncio.sleep(10)

                if self.running:
                    logger.warning("Lost connection to DAQ6510")
                    await self.daq_interface.disconnect()

            except (DeviceConnectionError, DeviceTimeoutError) as exc:
                logger.error(f"Failed to connect to DAQ6510: {exc}")
                await asyncio.sleep(5)  # Wait before retrying

    async def polling_loop(self):
        """Main polling loop for temperature measurements."""
        logger.info(f"Starting temperature polling loop (interval: {self.poll_interval}s, channels: {self.channels})")

        assert self.daq_interface is not None  # extra check and make mypy happy

        # The next lines are a way to calculate the sleep time between two measurements, this takes the time of the
        # measurement itself into account.
        def interval():
            next_time = time.perf_counter()
            while True:
                next_time += self.poll_interval
                yield max(next_time - time.perf_counter(), 0)

        g_interval = interval()

        while self.running and self.polling_active:
            try:
                if not await self.daq_interface.is_connected():
                    logger.warning("DAQ6510 not connected, skipping temperature reading")
                    await asyncio.sleep(5)
                    continue

                timestamp = datetime.datetime.now().isoformat()
                readings = {}

                # Read temperature from each channel
                for channel in self.channels:
                    try:
                        # temp = random.random()
                        temp = await self.daq_interface.get_measurement(channel)
                        readings[channel] = temp
                    except (DeviceConnectionError, DeviceTimeoutError, ValueError) as exc:
                        logger.error(f"Error reading channel {channel}: {exc}")
                        readings[channel] = None

                # Log the readings
                log_entry = {"timestamp": timestamp, "readings": readings}

                # Append to log file
                with open(self.log_file, "a") as fd:
                    fd.write(json.dumps(log_entry) + "\n")

                self._last_reading.update({"timestamp": timestamp, "readings": readings})

                logger.info(f"Temperature readings: {readings}")

            except Exception as exc:
                logger.exception(f"Error in polling loop: {exc}")

            finally:
                # Wait for next polling interval, we account for the time needed to perform the measurement.
                await asyncio.sleep(next(g_interval))

        logger.info("Temperature polling loop stopped")

    async def command_listener(self):
        """ZeroMQ command interface listener."""
        logger.info("Command listener started")

        while self.running:
            try:
                # Wait for next message
                message = await self.socket.recv_multipart()

                # Parse the message
                if len(message) < 3:
                    logger.warning(f"Received malformed message: {message}")
                    continue

                identity, empty, *payload = message

                try:
                    # Parse the command and parameters
                    command_data = json.loads(payload[0].decode("utf-8"))
                    command = command_data.get("command")
                    params = command_data.get("params", {})

                    logger.info(f"Received command: {command} from {identity}")

                    # Handle the command
                    if command in self.command_handlers:
                        response = await self.command_handlers[command](params)
                    else:
                        response = {"status": "error", "message": f"Unknown command: {command}"}

                except json.JSONDecodeError:
                    response = {"status": "error", "message": "Invalid JSON format"}
                except Exception as exc:
                    logger.exception(f"Error processing command: {exc}")
                    response = {"status": "error", "message": str(exc)}

                # Send response
                await self.socket.send_multipart([identity, b"", json.dumps(response).encode("utf-8")])

            except Exception as exc:
                logger.exception(f"Error in command listener: {exc}")
                await asyncio.sleep(1)

    async def _handle_start_polling(self, params: dict[str, Any]) -> dict[str, Any]:
        """Start temperature polling."""
        if not self.polling_active:
            self.polling_active = True

            # If channels provided, update them
            if "channels" in params:
                self.channels = params["channels"]

            # If interval provided, update it
            if "interval" in params:
                self.poll_interval = float(params["interval"])

            # Start polling loop
            polling_task = asyncio.create_task(self.polling_loop())

            # But we can add error handling for the task
            polling_task.add_done_callback(
                lambda t: logger.error(f"Polling loop ended unexpectedly: {t.exception()}") if t.exception() else None
            )

            return {
                "status": "ok",
                "message": f"Polling started with interval {self.poll_interval}s and channels {self.channels}",
            }
        else:
            return {"status": "ok", "message": "Polling already active"}

    async def _handle_stop_polling(self, params: dict[str, Any]) -> dict[str, Any]:
        """Stop temperature polling."""
        if self.polling_active:
            self.polling_active = False
            return {"status": "ok", "message": "Polling stopped"}
        else:
            return {"status": "ok", "message": "Polling already stopped"}

    async def _handle_set_interval(self, params: dict[str, Any]) -> dict[str, Any]:
        """Set polling interval."""
        if "interval" not in params:
            return {"status": "error", "message": "Missing required parameter: interval"}

        try:
            interval = float(params["interval"])
            if interval <= 0:
                return {"status": "error", "message": "Interval must be positive"}

            old_interval = self.poll_interval
            self.poll_interval = interval

            return {"status": "ok", "message": f"Polling interval changed from {old_interval}s to {interval}s"}
        except ValueError:
            return {"status": "error", "message": "Invalid interval format"}

    async def _handle_set_channels(self, params: dict[str, Any]) -> dict[str, Any]:
        """Set channels to monitor."""
        if "channels" not in params or not isinstance(params["channels"], list):
            return {"status": "error", "message": "Missing or invalid parameter: channels (should be a list)"}

        old_channels = self.channels.copy()
        self.channels = params["channels"]

        return {"status": "ok", "message": f"Monitoring channels changed from {old_channels} to {self.channels}"}

    async def _handle_get_last_reading(self, params: dict[str, Any]):
        return self._last_reading

    async def _handle_get_reading(self, params: dict[str, Any]):
        """Get a reading for the given channel(s)."""
        logger.info(f"GET_READING – {params = }")

        assert self.daq_interface is not None  # extra check and make mypy happy

        readings = {"status": "ok", "data": {}}

        for channel in params["channels"]:
            try:
                temp = await self.daq_interface.get_measurement(channel)
                readings["data"][channel] = temp
            except (DeviceConnectionError, DeviceTimeoutError, ValueError, RuntimeError) as exc:
                logger.error(f"Error reading channel {channel}: {exc}")
                readings["data"][channel] = None
                readings.update({"status": "error", "message": f"Error reading channel {channel}"})

        return readings

    async def _handle_get_status(self, params: dict[str, Any]) -> dict[str, Any]:
        """Get current service status."""
        connected = False
        try:
            if self.daq_interface:
                connected = await self.daq_interface.is_connected()
        except Exception:
            connected = False

        return {
            "status": "ok",
            "data": {
                "service_running": self.running,
                "polling_active": self.polling_active,
                "poll_interval": self.poll_interval,
                "channels": self.channels,
                "daq_connected": connected,
                "daq_hostname": self.daq_hostname,
                "daq_port": self.daq_port,
            },
        }

    async def _handle_shutdown(self, params: dict[str, Any]) -> dict[str, Any]:
        """Shutdown the service."""
        # Schedule shutdown after sending response
        _ = asyncio.create_task(self.shutdown())

        return {"status": "ok", "message": "Service shutting down"}

    async def shutdown(self):
        """Gracefully shut down the service."""
        logger.info("Shutting down DAQ Monitoring Service...")

        # Stop the main loops
        self.running = False
        self.polling_active = False

        # Disconnect DAQ
        try:
            logger.info("Disconnecting the DAQ6510...")
            if self.daq_interface and await self.daq_interface.is_connected():
                await self.daq_interface.disconnect()
        except Exception as exc:
            logger.error(f"Error disconnecting from DAQ: {exc}")

        # Close ZeroMQ socket
        try:
            logger.info("Closing ZeroMQ socket and terminate context...")
            self.socket.close()
            self.ctx.term()
        except Exception as exc:
            logger.error(f"Error closing ZeroMQ socket: {exc}")

        logger.info("Service shutdown complete")


class DAQMonitorClient:
    """A simple client for interacting with the DAQ Monitor Service."""

    def __init__(self, server_address: str = "localhost", port: int = DAQ_MON_CMD_PORT, timeout: float = 5.0):
        """Initialize the client.

        Args:
            server_address: Address of the monitoring service
            port: ZeroMQ port
            timeout: Command timeout in seconds
        """
        self.server_address = server_address
        self.port = port
        self.timeout = timeout

        self.ctx = zmq.Context().instance()
        self.socket = None

    def connect(self):
        """Connect to the DAQ Monitoring service."""
        self.socket = self.ctx.socket(zmq.DEALER)
        self.socket.connect(f"tcp://{self.server_address}:{self.port}")
        self.socket.setsockopt(zmq.RCVTIMEO, int(self.timeout * 1000))

    def disconnect(self):
        """Close the client connection."""
        if self.socket:
            self.socket.close(linger=100)
        self.ctx.term()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()
        if exc_type:
            logger.error(f"Caught {exc_type}: {exc_val}")

    def _send_command(self, command: str, params: dict[str, Any] | None = None) -> dict[str, Any]:
        """Send a command to the monitoring service.

        Args:
            command: Command name
            params: Optional command parameters

        Returns:
            Response from the service as a dictionary.
        """
        params = params or {}
        message = {"command": command, "params": params}

        assert self.socket is not None, "Client not connected. Call connect() first."

        try:
            self.socket.send_multipart([b"", json.dumps(message).encode("utf-8")])
            _, response_data = self.socket.recv_multipart()
            return json.loads(response_data.decode("utf-8"))
        except zmq.ZMQError as exc:
            return {"status": "error", "message": f"ZMQ error: {exc}"}
        except Exception as exc:
            return {"status": "error", "message": f"Error: {exc}"}

    def start_polling(self, channels: Optional[list[str]] = None, interval: Optional[float] = None) -> dict[str, Any]:
        """Start polling on specified channels.

        Args:
            channels: List of channels to monitor
            interval: Polling interval in seconds

        Returns:
            Response from the service
        """
        params = {}
        if channels is not None:
            params["channels"] = channels
        if interval is not None:
            params["interval"] = interval

        return self._send_command("START_POLLING", params)

    def stop_polling(self) -> dict[str, Any]:
        """Stop polling.

        Returns:
            Response from the service
        """
        return self._send_command("STOP_POLLING")

    def set_interval(self, interval: float) -> dict[str, Any]:
        """Set polling interval.

        Args:
            interval: New polling interval in seconds

        Returns:
            Response from the service
        """
        return self._send_command("SET_INTERVAL", {"interval": interval})

    def set_channels(self, channels: list[str]) -> dict[str, Any]:
        """Set channels to monitor.

        Args:
            channels: List of channel identifiers

        Returns:
            Response from the service
        """
        return self._send_command("SET_CHANNELS", {"channels": channels})

    def get_reading(self, channels: list[str]) -> dict[str, float]:
        """Get a reading from the given channel.

        Returns:
            A dictionary with the value of the measurement for the given channel.
        """
        return self._send_command("GET_READING", {"channels": channels})

    def get_last_reading(self) -> dict:
        return self._send_command("GET_LAST_READING")

    def get_status(self) -> dict[str, Any]:
        """Get current service status.

        To confirm the status is 'ok', check the response for the key 'status'.

        Returns:
            Status information as dictionary.
        """
        return self._send_command("GET_STATUS")

    def shutdown(self) -> dict[str, Any]:
        """Shutdown the service.

        Returns:
            Response from the service
        """
        return self._send_command("SHUTDOWN")


app = typer.Typer(name="daq6510_mon")


@app.command(cls=TyperAsyncCommand, name="monitor")
async def main(log_file: str = "temperature_readings.log"):
    """
    Start the DAQ6510 monitoring app in the background.
    """
    monitor = DAQ6510Monitor(
        daq_hostname=DAQ_DEV_HOST,
        daq_port=DAQ_DEV_PORT,
        zmq_port=DAQ_MON_CMD_PORT,
        log_file=log_file,
        channels=["101", "102"],
        poll_interval=10.0,
    )

    await monitor.start()


if __name__ == "__main__":
    asyncio.run(app())



__all__ = [
    "DataPoint",
    "TimeSeriesRepository",
    "define_metrics",
    "get_metrics_repo",
]

import datetime
from typing import Any
from typing import Optional
from typing import Protocol

import numpy as np
from prometheus_client import Gauge
from typing_extensions import Self

from egse.hk import TmDictionaryColumns
from egse.log import logger
from egse.plugin import load_plugins_fn
from egse.settings import Settings
from egse.setup import Setup
from egse.setup import SetupError
from egse.setup import load_setup
from egse.system import format_datetime
from egse.system import str_to_datetime

SITE_ID = Settings.load("SITE").ID

TimestampLike = str | int | float | datetime.datetime


def define_metrics(
    origin: str, dashboard: str | None = None, use_site: bool = False, setup: Optional[Setup] = None
) -> dict:
    """Creates a metrics dictionary from the telemetry dictionary.

    Read the metric names and their descriptions from the telemetry dictionary, and create Prometheus gauges based on
    this information.

    If `dashboard` is not provided, all telemetry parameters for the given origin will be returned.

    Args:
        origin: Storage mnemonics for the requested metrics
        dashboard: Restrict the metrics selection to those that are defined for the given dashboard. You can select
                   all dashboards with `dashboard='*'`.
        use_site: Indicate whether the prefixes of the new HK names are TH-specific
        setup: Setup.

    Returns: Dictionary with all Prometheus gauges for the given origin and dashboard.
    """

    setup = setup or load_setup()

    try:
        hk_info_table = setup.telemetry.dictionary
    except AttributeError:
        raise SetupError(
            "Version of the telemetry dictionary not specified in the current setup, "
            "expected to find it in setup.telemetry.dictionary. Cannot define metrics without telemetry dictionary."
        )

    hk_info_table = hk_info_table.replace(np.nan, "")

    storage_mnemonic = hk_info_table[TmDictionaryColumns.STORAGE_MNEMONIC].values
    hk_names = hk_info_table[TmDictionaryColumns.CORRECT_HK_NAMES].values
    descriptions = hk_info_table[TmDictionaryColumns.DESCRIPTION].values
    mon_screen = hk_info_table[TmDictionaryColumns.DASHBOARD].values

    condition = storage_mnemonic == origin.upper()
    if dashboard is not None:
        if dashboard == "*":
            extra_condition = mon_screen != ""
        else:
            extra_condition = mon_screen == dashboard.upper()
        condition = np.all((condition, extra_condition), axis=0)

    selection = np.where(condition)

    syn_names = hk_names[selection]
    descriptions = descriptions[selection]

    if not use_site:
        metrics = {}

        for syn_name, description in zip(syn_names, descriptions):
            try:
                metrics[syn_name] = Gauge(syn_name, description)
            except ValueError:
                logger.warning(f"ValueError for {syn_name}")

        return metrics

    th_prefix = f"G{SITE_ID}_"

    th_syn_names = []
    th_descriptions = []
    for syn_name, description in zip(syn_names, descriptions):
        if syn_name.startswith(th_prefix):
            th_syn_names.append(syn_name)
            th_descriptions.append(description)

    return {syn_name: Gauge(syn_name, description) for syn_name, description in zip(th_syn_names, th_descriptions)}


def update_metrics(metrics: dict, updates: dict):
    """Updates the metrics parameters with the values from the updates dictionary.

    Only the metrics parameters for which the names are keys in the given updates dict are actually updated. Other
    metrics remain untouched.

    The functions log a warning when the updates dict contains a name which is not known as a metrics parameter.

    Args:
        metrics: Metrics dictionary previously defined with the define_metrics function
        updates: Dictionary with key=metrics name and value is the to-be-updated value
    """

    for metric_name, value in updates.items():
        try:
            if value is None:
                metrics[metric_name].set(float("nan"))
            else:
                metrics[metric_name].set(float(value))
        except KeyError:
            logger.warning(f"Unknown metric name: {metric_name=}")


class PointLike(Protocol):
    @staticmethod
    def measurement(measurement_name: str) -> "PointLike": ...

    def tag(self, key, value) -> Self: ...

    def field(self, field, value) -> Self: ...

    def time(self, time: TimestampLike) -> Self: ...

    def as_dict(self) -> dict[str, Any]: ...


class DataPoint(PointLike):
    def __init__(self, measurement_name: str):
        self.measurement_name: str = measurement_name
        self.tags: dict[str, str] = {}
        self.fields: dict[str, Any] = {}
        self.timestamp: TimestampLike = format_datetime()

    def as_dict(self) -> dict[str, Any]:
        if isinstance(self.timestamp, str):
            timestamp = str_to_datetime(self.timestamp).timestamp()
        elif isinstance(self.timestamp, datetime.datetime):
            timestamp = self.timestamp.timestamp()
        else:
            timestamp = self.timestamp

        return {
            "measurement": self.measurement_name,
            "tags": self.tags,
            "fields": self.fields,
            "time": timestamp,
        }

    @staticmethod
    def measurement(measurement_name: str) -> "DataPoint":
        data_point = DataPoint(measurement_name)
        return data_point

    def tag(self, key, value):
        self.tags[key] = value
        return self

    def field(self, field, value):
        self.fields[field] = value
        return self

    def time(self, time: TimestampLike) -> Self:
        self.timestamp = time
        return self


class TimeSeriesRepository(Protocol):
    def connect(self) -> None: ...

    def ping(self) -> bool: ...

    def write(self, points: PointLike | dict | list[PointLike | dict]) -> None: ...

    def query(self, query_str: str, mode: str) -> Any: ...

    def get_table_names(self) -> list[str]: ...

    def get_column_names(self, table_name: str) -> list[str]: ...

    def get_values_last_hours(self, table_name: str, column_name: str, hours: int, mode: str) -> Any: ...

    def get_values_in_range(
        self, table_name: str, column_name: str, start_time: str, end_time: str, mode: str
    ) -> Any: ...

    def close(self) -> None: ...


def get_metrics_repo(plugin_name: str, config: dict[str, Any]) -> TimeSeriesRepository:
    """
    Create a TimeSeriesRepository instance from a plugin.

    Args:
        plugin_name: Name of the plugin (without .py extension)
        config: Configuration parameters for the repository

    Returns:
        Configured TimeSeriesRepository instance.

    Raises:
        ModuleNotFoundError: If plugin not found
        NotImplementedError: If plugin is missing get_repository_class()
    """

    package_name = "egse.plugins"
    plugins = load_plugins_fn(f"{plugin_name}.py", package_name)

    if plugin_name in plugins:
        plugin = plugins[plugin_name]
        if hasattr(plugin, "get_repository_class"):
            repo_class = plugin.get_repository_class()
        else:
            raise NotImplementedError(f"Missing 'get_repository_class()` function in metrics plugin {plugin_name}.")
        return repo_class(**config)
    else:
        raise ModuleNotFoundError(f"No plugin found for {plugin_name} in {package_name}.")



import logging
import time
from pathlib import Path

import pytest

from egse.env import get_log_file_location
from egse.logger import create_new_zmq_logger
from egse.logger import egse_logger
from egse.logger import get_log_file_name
from egse.logger import send_request
from egse.process import is_process_running
from egse.system import read_last_lines


@pytest.mark.skipif(
    bool(is_process_running(items=["log_cs"])),
    reason="This test starts its own logging server, log_cs can not be running.",
)
def test_logging_messages_of_different_levels(setup_log_service):
    # The egse logger doesn't propagate messages to parent loggers, so we
    # have to add the caplog handler in order to capture logging messages for this test.
    # egse_logger.addHandler(caplog.handler)
    # egse_logger.setLevel(logging.DEBUG)

    egse_logger.debug("This is a DEBUG message.")
    egse_logger.info("This is a INFO message.")
    egse_logger.warning("This is a WARNING message.")
    egse_logger.error("This is a ERROR message.")
    egse_logger.critical("This is a CRITICAL message.")

    time.sleep(1.0)  # give some time for the log messages to be written to file

    status = send_request("status")

    log_location = get_log_file_location() + "/" + get_log_file_name()
    log_location = status.get("file_logger_location", log_location)

    lines = read_last_lines(filename=Path(log_location), num_lines=10)

    # FIXME:
    #    In my setup, lines is [], so nothing has been read from the log file. When I inspect
    #    the log file after the test, the expected line is there!
    #    -> check if the log_cs service is running and writing to the expected location.

    print(f"{log_location = }, {get_log_file_name()=}, {len(lines)=}")
    for line in lines:
        print(line)

    # The DEBUG message should be in the log file that was created by the log_cs, because
    # the log level is set to DEBUG for the log-file.

    assert any([True if "This is a DEBUG message." in x else False for x in lines])


def test_logging_exception(caplog):
    try:
        raise ValueError("incorrect value entered.")
    except ValueError:
        egse_logger.exception("Reporting a ValueError")

    assert "incorrect value entered" in caplog.text
    assert "Reporting a ValueError" in caplog.text


def test_logging_error(caplog):
    try:
        raise ValueError("incorrect value entered.")
    except ValueError:
        egse_logger.error("Reporting a ValueError")
        egse_logger.error("Reporting a ValueError with exc_info", exc_info=True)

    assert "incorrect value entered" in caplog.text
    assert "with exc_info" in caplog.text


def test_create_new_zmq_logger(caplog):
    print()

    camtest_logger = create_new_zmq_logger("camtest")

    camtest_logger.info("First message with ZeroMQ handler in camtest logger")

    assert "camtest:test_logger" in caplog.text
    assert "ZeroMQ" in caplog.text

    print(f"{caplog.text = }")

    caplog.clear()

    logger = logging.getLogger("camtest.sub_level")

    logger.info("Message from sub_level logger should be categorised under camtest_logger")

    assert "camtest.sub_level:test_logger" in caplog.text
    assert "categorised" in caplog.text

    # See what happens if we call the function twice with the same logger

    caplog.clear()

    camtest_logger = create_new_zmq_logger("camtest")

    # If the following message appears twice in the general.log logfile then a second handler
    # was created by the create_new_zmq_logger function.

    camtest_logger.info("Created the zmq handler twice?")

    lines = caplog.text.split("\n")
    lines = [line for line in lines if line.strip()]  # filter out empty lines

    assert len(lines) == 1



"""
The Reference Frame Model is a self-consistent graph of reference frames that are connected by
reference and/or by explicit links.

The ReferenceFrameModel class will keep the model consistent and self-contained.

Functionality:

* Initialization from a list of ReferenceFrames.
* Serialization into a dictionary that can be saved into a YAML file.
* Manipulation of the model
    * Move a reference frame (translation, rotation) with respect to another reference frame
        * Absolute movement, center of rotation either local or other
        * Relative Movement, center of rotation either local or other
    * Move a reference frame (translation, rotation) with respect to itself
        * Absolute movement
        * Relative Movement
    * Change the definition of a reference frame in the model
* Inspection of the model
    * Get the definition of a reference frame (what should this be? only translation & rotation?)
    * Get position of a reference frame
    * Get the position of a point in a 'target' reference frame, but defined in a 'source'
      reference frame
    * Get a string representation of the model.
    * Find inconsistencies in the model
    * What other information do we need?
        * find the path from one reference frame to another reference frame?
        * find all reference frames that are affected by a movement or redefinition of a
          reference frame?
        * ...

"""

import logging
from typing import Dict
from typing import List

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.transforms import ScaledTranslation
from mpl_toolkits.mplot3d import Axes3D

import egse.coordinates.transform3d_addon as t3add
from egse.coordinates import dict_to_ref_model
from egse.coordinates import ref_model_to_dict
from egse.setup import NavigableDict

LOGGER = logging.getLogger(__name__)

# TODO : HANDLING "moving_in_ref" (obusr)  in move_absolute_ext and move_relative_ext
#        should it be added to the model temporarily ??
#        after the move : remove the link and delete that frame
#
# Priority 1
#   * access methods to allow for things like is_avoidance_ok(hexsim.cs_user, hexsim.cs_object,
#     setup=setup, verbose=True)
#
#  Priority 2
#   * Move name handling from ReferenceFrame to here (if necessary)


class ReferenceFrameModel:
    """
    A ReferenceFrameModel is a collection of reference frames that are linked to each other to
    form a Graph.
    """

    _ROT_CONFIG_DEFAULT = "sxyz"
    _ACTIVE_DEFAULT = True
    _DEGREES_DEFAULT = True

    def __init__(
        self,
        model: Dict | List = None,
        rotation_config: str = _ROT_CONFIG_DEFAULT,
        use_degrees: bool = _DEGREES_DEFAULT,
        use_active_movements: bool = _ACTIVE_DEFAULT,
    ):
        """Initialisation of a reference frame model.

        Args:
            model (Dict | List[ReferenceFrame]): List or a dictionary of reference frames that make up the model.
            rotation_config (str): Order in which the rotation about the three axes are chained.
            use_degrees (bool): Indicates whether the rotation angles are specified in degrees, rather than radians.
            use_active_movements (bool): Indicates if the rotation is active (object rotates IN a fixed coordinate
                                         system) or passive (coordinate system rotates AROUND a fixed object).  Even if
                                         two angles are zero, the match between angle orders and rot_config is still
                                         critical.
        """

        self._use_degrees = use_degrees
        self._use_active_movements = use_active_movements
        self._rot_config = rotation_config

        # Keep a dictionary with all reference frames that are part of the model. The keys shall
        # be the name of the reference frame. When the model passed is empty, create only a
        # master reference frame.

        if isinstance(model, (dict, list)):
            self._model = self.deserialize(model)
        else:
            self._model = NavigableDict({})

    def __str__(self) -> str:
        """Returns a printable string representation of the reference frame.

        Returns:
            Printable string representation of the reference frame.
        """

        return self._model.pretty_str()

    def __len__(self) -> int:
        """Returns the number of reference frames in the model."""

        return len(self._model)

    def __contains__(self, item) -> bool:
        """Checks whether the given item is present in the model.

        Args:
            item: Item for which to check whether it's present in the model.

        Returns:
            True if the given item is present in the model; False otherwise.
        """

        return item in self._model

    def __iter__(self):
        """Returns an iterator over the reference frames in the model."""

        return iter(self._model.values())

    def summary(self) -> str:
        """Returns a summary of the model.

        Returns:
            String summary of the model.
        """

        result = f"Number of frames: {len(self)}\n"

        for reference_frame in self:
            result += f"{reference_frame.name:>10}[{reference_frame.ref.name}]  ---  {[link.name for link in reference_frame.linked_to]}\n"

        return result

    @staticmethod
    def deserialize(model_dict: dict) -> NavigableDict:
        """De-serialisation of the model.

        De-serialisation means you take a serialised representation of a reference frames model and turn it into a
        dictionary containing all the reference frames with their links and references.

        Args:
            model_dict (dict): Dictionary of serialised reference frames.

        Returns:
            Dictionary of reference frames that make up a model.

        """
        return dict_to_ref_model(model_dict)

    def serialize(self) -> NavigableDict:
        """Serialisation of the model.


        Serialisation of the model by serialising each of the reference frames into an object that can easily be saved
        to a YAML or a JSON file.

        Returns:
            Dictionary with all the serialised reference framed.
        """

        return ref_model_to_dict(self._model)

    def add_master_frame(self) -> None:
        """Adds the master reference frame to the model."""

        # TODO: First check if there is not already a Master frame in the model

        from egse.coordinates.reference_frame import ReferenceFrame

        self._model["Master"] = ReferenceFrame.create_master()

    def add_frame(
        self,
        name: str,
        *,
        translation: np.ndarray = None,
        rotation: np.ndarray = None,
        transformation: np.ndarray = None,
        ref: str,
    ) -> None:
        """Adds a reference frame to the model.

        Args:
            name (str): Name of the reference frame to add to the model.  Only this parameter can be positional.  This
                        will serve as the identifier for the reference frame in the model.
            translation (np.ndarray): Translation vector. Ignored when `transformation` is given.
            rotation (np.ndarray: Rotation vector. Ignored when `transformation` is given.
            transformation (np.ndarray): Transformation matrix.
            ref (str): Name of the reference frame that is a reference for the new reference frame, i.e. the new
                             reference frame is defined w.r.t. this one.
        """

        from egse.coordinates.reference_frame import ReferenceFrame

        if name in self._model:
            raise KeyError("A reference frame with the name '{name} already exists in the model.")

        ref = self._model[ref]

        if transformation:
            self._model[name] = ReferenceFrame(transformation, ref=ref, name=name, rotation_config=self._rot_config)
        else:
            self._model[name] = ReferenceFrame.from_translation_rotation(
                translation,
                rotation,
                name=name,
                ref=ref,
                rotation_config=self._rot_config,
                degrees=self._use_degrees,
                active=self._use_active_movements,
            )

    def remove_frame(self, name: str):
        """Deletes the given reference frame from the model.

        If the reference frame doesn't exist in the model, a warning message is logged.

        Args:
            name (str): Name of the reference frame to remove.
        """

        if name in self._model:
            from egse.coordinates.reference_frame import ReferenceFrame

            frame: ReferenceFrame = self._model[name]

            # We need to get the links out in a list because the frame.remove_link() method deletes
            # frames from the linked_to dictionary and that is not allowed in a for loop.

            links = [linked_frame for linked_frame in frame.linked_to]
            for link in links:
                frame.remove_link(link)

            del self._model[name]
        else:
            LOGGER.warning(f"You tried to remove a non-existing reference frame '{name}' from the model.")

    def get_frame(self, name: str):
        """Returns the reference frame with the given name.

        Use this function with care since this breaks encapsulation and may lead to an inconsistent model when the frame
        is changed outside the scope of the reference model.

         Args:
             name (str): Name of the requested reference frame.

         Returns:
             Reference frame with the given name.
        """

        return self._model[name]

    def add_link(self, source: str, target: str):
        """Adds a link between the two given reference frames of the model.

        All links are bi-directional.

        Args:
            source (args): Name of the source reference frame.
            target (args): Name of the target reference frame.
        """

        if source not in self._model:
            raise KeyError(f"There is no reference frame with the name '{source} in the model.")
        if target not in self._model:
            raise KeyError(f"There is no reference frame with the name '{target} in the model.")

        source = self._model[source]
        target = self._model[target]

        source.add_link(target)

    def remove_link(self, source: str, target: str):
        """Removes a link between two reference frames.

        All links are bi-directional and this method removes both links.

        Args:
            source (args): Name of the source reference frame.
            target (args): Name of the target reference frame.
        """

        if source not in self._model:
            raise KeyError(f"There is no reference frame with the name '{source} in the model.")
        if target not in self._model:
            raise KeyError(f"There is no reference frame with the name '{target} in the model.")

        source = self._model[source]
        target = self._model[target]

        source.remove_link(target)

    def move_absolute_self(
        self, name: str, translation: np.ndarray, rotation: np.ndarray, degrees: bool = _DEGREES_DEFAULT
    ) -> None:
        """Applies an absolute movement to the given reference frame.

        Applies an absolute movement to the given reference frame such that it occupies a given absolute position w.r.t.
        "frame_ref" after the movement.

        There is no hexapod equivalent.

        Args:
            name (str): Name of the reference frame to move.
            translation (np.ndarray): Translation vector.
            rotation (np.ndarray): Rotation vector.
            degrees (bool): Indicates whether the rotation angles are specified in degrees, rather than radians.

        Args:
            name (str): the name of the reference frame to move
        """

        name = self._model[name]
        name.set_translation_rotation(
            translation,
            rotation,
            rotation_config=self._rot_config,
            active=self._use_active_movements,
            degrees=degrees,
            preserve_links=True,
        )

    def move_absolute_in_other(
        self, frame: str, other: str, translation: np.ndarray, rotation: np.ndarray, degrees: bool = _DEGREES_DEFAULT
    ):
        """Applies an absolute movement to the given reference frame in another reference frame.

        Apply an absolute movement to the ReferenceFrame "frame", such that it occupies a given absolute position
        w.r.t. "other" after the movement.

        Hexapod equivalent: PunaSimulator.move_absolute, setting `hexobj` w.r.t. `hexusr`.

        Args:
            frame (str): Name of the reference frame to move.
            other (str): Name of the other reference frame.
            translation (np.ndarray): Translation vector.
            rotation (np.ndarray): Rotation vector.
            degrees (bool): Indicates whether the rotation angles are specified in degrees, rather than radians.
        """

        # TODO:
        #   There can not be a link between frame and other, not direct and not indirect.
        #   So, with A-link-B-link-C-link-D, we can not do move_absolute_in_other('A', 'D', ...)

        frame = self._model[frame]
        other = self._model[other]

        transformation = other.get_active_transformation_to(frame)

        from egse.coordinates.reference_frame import ReferenceFrame

        moving_in_other = ReferenceFrame(
            transformation, ref=other, name="moving_in_other", rotation_config=self._rot_config
        )

        moving_in_other.add_link(frame)

        moving_in_other.set_translation_rotation(
            translation,
            rotation,
            rotation_config=self._rot_config,
            active=self._use_active_movements,
            degrees=degrees,
            preserve_links=True,
        )

        moving_in_other.remove_link(frame)

        del moving_in_other

    def move_relative_self(
        self, frame: str, translation: np.ndarray, rotation: np.ndarray, degrees: bool = _DEGREES_DEFAULT
    ):
        """Applies a relative movement to the given reference frame.

        It is assumed that the movement is expressed in that same reference frame.

        Hexapod equivalent: PunaSimulator.move_relative_object

        Args:
            frame (str): Name of the reference frame to move.
            translation (np.ndarray): Translation vector.
            rotation (np.ndarray): Rotation vector.
            degrees (bool): Indicates whether the rotation angles are specified in degrees, rather than radians.
        """

        frame = self._model[frame]
        frame.apply_translation_rotation(
            translation,
            rotation,
            rotation_config=self._rot_config,
            active=self._use_active_movements,
            degrees=degrees,
            preserve_links=True,
        )

    def move_relative_other(
        self, frame: str, other: str, translation: np.ndarray, rotation: np.ndarray, degrees: bool = _DEGREES_DEFAULT
    ):
        """Applies a relative movement to the given reference frame.

        The movement is expressed w.r.t. the axes of another frame.  The centre of rotation is the origin of the
        that other reference frame.

        There is no hexapod equivalent.

        Args:
            frame (str): Name of the reference frame to move.
            other (str): Name of the reference frame in which the movements have been defined.
            translation (np.ndarray): Translation vector.
            rotation (np.ndarray): Rotation vector.
            degrees (bool): Indicates whether the rotation angles are specified in degrees, rather than radians.
        """

        # TODO:
        #   There can not be a link between frame and other, not direct and not indirect.
        #   So, with A-link-B-link-C-link-D, we can not do move_absolute_in_other('A', 'D', ...)

        frame = self._model[frame]
        other = self._model[other]

        transformation = frame.get_active_transformation_to(other)

        from egse.coordinates.reference_frame import ReferenceFrame

        moving_in_other = ReferenceFrame(
            transformation, ref=other, name="moving_in_other", rotation_config=self._rot_config
        )

        moving_in_other.add_link(frame)

        moving_in_other.apply_translation_rotation(
            translation,
            rotation,
            rotation_config=self._rot_config,
            active=self._use_active_movements,
            degrees=degrees,
            preserve_links=True,
        )

        moving_in_other.remove_link(frame)

        del moving_in_other  # not need as local scope

    def move_relative_other_local(
        self, frame: str, other: str, translation: np.ndarray, rotation: np.ndarray, degrees: bool = _DEGREES_DEFAULT
    ):
        """Applies a relative movement to the given reference frame.

        The movement is expressed w.r.t. the axes of another reference frame.  The centre of rotation is the origin of
        that other reference frame.

        Hexapod equivalent: PunaSimulator.move_relative_user

        Args:
            frame (str): Name of the reference frame to move.
            other (str): Name of the reference frame in which the movements have been defined.
            translation (np.ndarray): Translation vector.
            rotation (np.ndarray): Rotation vector.
            degrees (bool): Indicates whether the rotation angles are specified in degrees, rather than radians.
        """

        # TODO:
        #   There can not be a link between frame and other, not direct and not indirect.
        #   So, with A-link-B-link-C-link-D, we can not do move_absolute_in_other('A', 'D', ...)

        frame = self._model[frame]
        other = self._model[other]

        # Represent the requested movement
        # De-rotation of MOVING -> REF  (align frame_moving axes on those of frame_ref)

        derotation = frame.get_active_transformation_to(other)
        derotation[:3, 3] = [0, 0, 0]

        # Reverse rotation (record inverse rotation, to restore the frame in the end)

        rerotation = derotation.T

        # Requested translation matrix  (already expressed wrt frame_ref)

        translation_ = np.identity(4)
        translation_[:3, 3] = translation

        # Requested rotation matrix (already expressed wrt frame_ref)

        zeros = [0, 0, 0]
        rotation_ = t3add.translation_rotation_to_transformation(
            zeros, rotation, rotation_config=self._rot_config, degrees=degrees
        )

        # All translations and rotations are applied to frame_moving
        # -> a. Need for "de-rotation" before applying the translation
        #    b. The centre of rotation is always the origin of frame_moving
        # 1. Rotate frame_moving to align it with frame_ref (i.e. render their axes parallel)
        # 2. Apply the translation in this frame
        # 3. Restore the original orientation of the moving frame
        # 4. Apply the requested rotation

        transformation = derotation @ translation_ @ rerotation @ rotation_

        # Apply the requested movement

        frame.apply_transformation(transformation, preserve_links=True)


def plot_ref_model(model: ReferenceFrameModel) -> None:
    """Plots the xz-plane of the given reference frame model."""

    # figsize is in inch, 6 inch = 15.24 cm, 5 inch = 12.7 cm

    fig = plt.figure(figsize=(6, 5), dpi=100)

    ax = fig.add_subplot(1, 1, 1)

    # Set axes limits in data coordinates

    ax.set_xlim(-10, 10)
    ax.set_ylim(-10, 10)
    ax.set_xticks(range(-10, 11, 2))
    ax.set_yticks(range(-10, 11, 2))
    ax.grid(True)

    for frame in model:
        draw_frame(ax, frame, plane="xz")

    plt.show()


def plot_ref_model_3d(model: ReferenceFrameModel) -> None:
    """Plots the given reference frame model in 3D."""

    fig = plt.figure(figsize=(8, 8), dpi=100)
    ax = Axes3D(fig)
    # ax.set_box_aspect([1, 1, 1])

    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")

    def get_fix_mins_maxs(mins, maxs):
        deltas = (maxs - mins) / 12.0
        mins = mins + deltas / 4.0
        maxs = maxs - deltas / 4.0

        return [mins, maxs]

    min_ = -8
    max_ = 8
    minmax = get_fix_mins_maxs(min_, max_)

    # ax.set_xticks(range(min_, max_, 2))
    # ax.set_yticks(range(min_, max_, 2))
    # ax.set_zticks(range(min_, max_, 2))

    ax.set_xlim(minmax)
    ax.set_ylim(minmax)
    ax.set_zlim(minmax)

    delta = 0.1
    ax.scatter(
        [min_ + delta, max_ - delta],
        [min_ + delta, max_ - delta],
        [min_ + delta, max_ - delta],
        color="k",
        marker=".",
    )

    for frame in model:
        draw_frame_3d(ax, frame)

    # ax.set_proj_type('ortho')
    ax.set_proj_type("persp")

    set_axes_equal(ax)
    plt.show()


def set_axes_equal(ax: plt.Axes) -> None:
    """Sets 3D plot axes to equal scale.

    Make axes of 3D plot have equal scale so that spheres appear as spheres and cubes as cubes.  Required since
    `ax.axis('equal')` and `ax.set_aspect('equal')` don't work on 3D.

    The aspect rati0 of the plots is not equal by default.  This solution was given in Stack Overflow:
    https://stackoverflow.com/a/63625222/4609203
    """

    limits = np.array([ax.get_xlim3d(), ax.get_ylim3d(), ax.get_zlim3d()])
    origin = np.mean(limits, axis=1)
    radius = 0.5 * np.max(np.abs(limits[:, 1] - limits[:, 0]))
    x, y, z = origin
    ax.set_xlim3d([x - radius, x + radius])
    ax.set_ylim3d([y - radius, y + radius])
    ax.set_zlim3d([z - radius, z + radius])


def draw_frame_3d(ax: Axes3D, frame, **kwargs) -> None:
    """Draws the given frame in 3D.

    Args:
        ax (Axes3D): Axis to draw the frame in.
        frame (ReferenceFrame): Reference frame to draw.
        **kwargs: Keyword arguments to pass to the quiver function.
    """

    master = frame.find_master()

    f0 = frame.get_origin()
    fx = frame.get_axis("x", name="fx")
    fy = frame.get_axis("y", name="fy")
    fz = frame.get_axis("z", name="fz")
    f0m = f0.express_in(master)[:3]
    fxm = fx.express_in(master)[:3]
    fym = fy.express_in(master)[:3]
    fzm = fz.express_in(master)[:3]

    # Origin of the x, y. and z vectors (x = the 'x' coordinates of the origin of all 3 vectors)
    # Every vector independently (-> plot in different colours)

    x, y, z = np.array([f0m[0]]), np.array([f0m[1]]), np.array([f0m[2]])

    # Orientation of the x, y, and z vectors

    vecxx, vecyx, veczx = (
        np.array([fxm[0] - f0m[0]]),
        np.array([fym[0] - f0m[0]]),
        np.array([fzm[0] - f0m[0]]),
    )
    vecxy, vecyy, veczy = (
        np.array([fxm[1] - f0m[1]]),
        np.array([fym[1] - f0m[1]]),
        np.array([fzm[1] - f0m[1]]),
    )
    vecxz, vecyz, veczz = (
        np.array([fxm[2] - f0m[2]]),
        np.array([fym[2] - f0m[2]]),
        np.array([fzm[2] - f0m[2]]),
    )

    kwargs.setdefault("length", 2)
    kwargs.setdefault("normalize", True)

    ax.quiver(x, y, z, vecxx, vecxy, vecxz, color="r", **kwargs)
    ax.quiver(x, y, z, vecyx, vecyy, vecyz, color="g", **kwargs)
    ax.quiver(x, y, z, veczx, veczy, veczz, color="b", **kwargs)

    offset = 0.1
    ax.text(f0m[0] + offset, f0m[1] + offset, f0m[2] + offset, frame.name)


def draw_frame(ax: plt.Axes, reference_frame, plane="xz", default_axis_length: int = 100) -> None:
    """Draws the given plane from the given reference frame in the given axis.

    Args:
        ax (plt.Axes): Axis to draw the plane in.
        reference_frame (ReferenceFrame): Reference frame to draw.
        plane (str, optional): Kind of plane to draw.  Must be in ["xy", "yz", "zx"].
        default_axis_length (int): Axis length.
    """

    fig = ax.get_figure()

    # FC : Figure coordinates (pixels)
    # NFC : Normalized figure coordinates (0 → 1)
    # DC : Data coordinates (data units)
    # NDC : Normalized data coordinates (0 → 1)

    dc2fc = ax.transData.transform
    fc2dc = ax.transData.inverted().transform
    fc2ndc = ax.transAxes.inverted().transform

    def dc2ndc(x):  # better than defining and assigning a lambda function
        return fc2ndc(dc2fc(x))

    x_idx, y_idx = {"xz": (0, 2), "xy": (0, 1), "yz": (1, 2)}[plane]

    # Draw the origin

    origin = reference_frame.get_origin()
    origin_in_master = origin.express_in(reference_frame.find_master())

    ax.scatter([origin_in_master[x_idx]], [origin_in_master[y_idx]], color="k")

    # Draw the axis

    origin_dc = np.array([[origin_in_master[x_idx], origin_in_master[y_idx]]])

    point = dc2fc(origin_dc[0])
    point[0] += default_axis_length
    target_dc = np.append(origin_dc, [fc2dc(point)], axis=0)

    ax.plot(target_dc[:, 0], target_dc[:, 1], color="k")

    point = dc2fc(origin_dc[0])
    point[1] += default_axis_length
    target_dc = np.append(origin_dc, [fc2dc(point)], axis=0)

    ax.plot(target_dc[:, 0], target_dc[:, 1], color="k")

    # Draw the axes label

    dx, dy = 10 / fig.dpi, 10 / fig.dpi
    offset = ScaledTranslation(dx, dy, fig.dpi_scale_trans)
    point = dc2ndc(origin_dc[0])
    plt.text(point[0], point[1], reference_frame.name, transform=ax.transAxes + offset)


def define_the_initial_setup() -> ReferenceFrameModel:
    """Defines the initial setup of the reference frame model.

    Returns:
        Reference frame model with the initial setup.
    """

    model = ReferenceFrameModel()

    model.add_master_frame()
    model.add_frame("A", translation=[2, 2, 2], rotation=[0, 0, 0], ref="Master")
    model.add_frame("B", translation=[-2, 2, 2], rotation=[0, 0, 0], ref="Master")
    model.add_frame("C", translation=[2, 2, 5], rotation=[0, 0, 0], ref="A")
    model.add_frame("D", translation=[2, 2, 2], rotation=[0, 0, 0], ref="B")

    model.add_link("A", "B")
    model.add_link("B", "C")

    print(model.serialize())
    plot_ref_model_3d(model)

    return model


def get_vectors(reference_frame_1, reference_frame_2, model: ReferenceFrameModel) -> tuple[np.ndarray, np.ndarray]:
    """Returns the translation and rotation vectors for the active transformation for the 1st reference frame to the 2nd.

    Args:
        reference_frame_1 (str): Name of the reference frame to get the active transformation from.
        reference_frame_2 (str): Name of the reference frame to get the active transformation to.
        model (ReferenceFrameModel): Model containing the reference frames with the given names.

    Returns:
        Translation and rotation vectors from the active transformation from the 1st reference frame to the 2nd.
    """

    return model.get_frame(reference_frame_1).get_active_translation_rotation_vectors_to(
        model.get_frame(reference_frame_2)
    )


def print_vectors(reference_frame_1: str, reference_frame_2: str, model: ReferenceFrameModel) -> None:
    """Prints the translation and rotation vectors for the active transformation for the 1st reference frame to the 2nd.

    Args:
        reference_frame_1 (str): Name of the reference frame to get the active transformation from.
        reference_frame_2 (str): Name of the reference frame to get the active transformation to.
        model (ReferenceFrameModel): Model containing the reference frames with the given names.
    """

    trans, rot = model.get_frame(reference_frame_1).get_active_translation_rotation_vectors_to(
        model.get_frame(reference_frame_2)
    )
    print(
        f"{reference_frame_1:8s} -> {reference_frame_2:8s} : Trans [{trans[0]:11.4e}, {trans[1]:11.4e}, {trans[2]:11.4e}]    Rot [{rot[0]:11.4e}, {rot[1]:11.4e}, {rot[2]:11.4e}]"
    )
    return


if __name__ == "__main__":
    logging.basicConfig(level=20)

    model = define_the_initial_setup()

    print(model.summary())

    print("\nMove frame 'A', frames 'B' and 'C' move with it.\n")
    model.move_absolute_self("A", [1, 1, 3], [0, 0, 45])
    print(model.serialize())
    plot_ref_model_3d(model)

    model = define_the_initial_setup()

    print("\nMove frame 'B' with respect to 'Master, frames 'A' and 'C' move with it.\n")
    model.move_absolute_in_other("B", "Master", [1, 1, -1], [0, 0, 0])
    print(model.serialize())
    plot_ref_model_3d(model)

    model = define_the_initial_setup()

    print("\nMove frame 'D' relative to itself, turn 45º\n")
    model.move_relative_self("D", [0, 0, 0], [45, 0, 0])
    print(model.serialize())
    plot_ref_model_3d(model)

    model = define_the_initial_setup()

    print("\nMove frame 'D' relative to 'A', turn 45º around origin of 'A'\n")
    model.move_relative_other("D", "A", [0, 0, 0], [0, 45, 0])
    print(model.serialize())
    plot_ref_model_3d(model)

    model = define_the_initial_setup()

    print("\nMove frame 'D' relative to 'A', turn 45º around origin of 'D'\n")
    model.move_relative_other_local("D", "A", [0, 0, 0], [0, 45, 0])
    print(model.serialize())
    plot_ref_model_3d(model)

    model = define_the_initial_setup()



import logging
import subprocess
import textwrap
from pathlib import Path
from typing import Union

from egse.command import ClientServerCommand
from egse.connect import get_endpoint
from egse.control import is_control_server_active
from egse.decorators import dynamic_interface
from egse.plugin import entry_points
from egse.process import SubProcess
from egse.proxy import Proxy
from egse.registry.client import RegistryClient
from egse.settings import Settings
from egse.setup import Setup
from egse.setup import load_setup
from egse.storage import is_storage_manager_active

HERE = Path(__file__).parent
LOGGER = logging.getLogger("egse.procman")

settings = Settings.load("Process Manager Control Server")
COMMAND_SETTINGS = Settings.load(location=HERE, filename="procman.yaml")

PROXY_TIMEOUT = 10.0  # don't wait longer than 10s

PROCESS_NAME = settings.get("PROCESS_NAME", "pm_cs")
SERVICE_TYPE = settings.get("SERVICE_TYPE", "pm_cs")
PROTOCOL = settings.get("PROTOCOL", "tcp")
HOSTNAME = settings.get("HOSTNAME", "localhost")
COMMANDING_PORT = settings.get("COMMANDING_PORT", 0)  # dynamically assigned by the system if 0
SERVICE_PORT = settings.get("SERVICE_PORT", 0)
MONITORING_PORT = settings.get("MONITORING_PORT", 0)
STORAGE_MNEMONIC = settings.get("STORAGE_MNEMONIC", "PM")


def is_process_manager_active(timeout: float = 0.5) -> bool:
    """Checks if the Process Manager Control Server is active.

    To check whether the Process Manager is active, a "Ping" command is sent.  If a "Pong" reply is received before
    timeout, that means that the Control Server is active (and True will be returned).  If no reply is received before
    timeout or if the reply is not "Pong", the Control Server is inactive (and False will be returned).

    Args:
        - timeout (float): Timeout when waiting for a reply [s] from the Control Server

    Returns: True if the Process Manager Control Server is active; False otherwise.
    """

    with RegistryClient() as client:
        endpoint = client.get_endpoint(settings.SERVICE_TYPE)

    if endpoint is None:
        return False

    return is_control_server_active(endpoint, timeout)


def get_status() -> str:
    """Returns a string representing the status of the Process Manager.

    Returns: String representation of the status of the Process Manager.
    """

    try:
        with ProcessManagerProxy() as sm:
            text = textwrap.dedent(
                f"""\
                Process Manager:
                    Status: [green]active[/]
                    Hostname: {sm.get_ip_address()}
                    Monitoring port: {sm.get_monitoring_port()}
                    Commanding port: {sm.get_commanding_port()}
                    Service port: {sm.get_service_port()}
                """
            )
        return text

    except ConnectionError as exc:
        return f"Process Manager Status: [red]not active[/] ({exc})"


class ProcessCommand:
    def __init__(self, device_id: str, device_args: Union[list, None], cgse_cmd: str):
        """Initialisation of a process command for a device Control Server.

        Args:
            device_id (str): Device identifier.
            device_args (list): Device arguments.
            cgse_cmd (str): CGSE command to start/stop the Control Server or to query its status.
        """

        self._device_id = device_id
        self._cgse_cmd = cgse_cmd
        self._device_args = device_args

    @property
    def device_id(self) -> str:
        """Returns the device identifier.

        Returns: Device identifier
        """

        return self._device_id

    @property
    def device_args(self) -> Union[list, None]:
        """Returns the device arguments.

        Returns: Device arguments
        """

        return self._device_args

    @property
    def cmd(self):
        raise NotImplementedError


class StartCommand(ProcessCommand):
    """Command to start the Control Server for a device."""

    def __init__(self, device_id: str, cgse_cmd: str, device_args: Union[list, None], is_simulator_mode: bool = False):
        """Initialisation of a start command for a device Control Server.

        Args:
            device_id (str): Device identifier.
            cgse_cmd (str): CGSE command to start/stop the Control Server or to query its status.
            device_args (Union[list, None]): Device arguments.
            is_simulator_mode (bool): Whether to start the Control Server in simulator mode rather than operational mode.
        """

        super().__init__(device_id, device_args, cgse_cmd)

        self._is_simulator_mode = is_simulator_mode

    @property
    def cmd(self) -> str:
        """Returns the full CGSE command to start the Control Server.

        Returns: Full CGSE command to start the Control Server.
        """

        cmd = f"{self._cgse_cmd} start"

        if self.device_args:
            cmd = f"{cmd} {self.device_args[0]}"
        if self.simulator_mode:
            cmd = f"{cmd} --sim"

        return cmd

    @property
    def simulator_mode(self) -> bool:
        """Checks whether the Control Server should be started in simulator mode rather than operational mode.

        Returns: True if the Control Server should be started in simulator mode; False otherwise.
        """

        return self._is_simulator_mode


class StopCommand(ProcessCommand):
    """Command to stop the Control Server for a device."""

    def __init__(self, device_id: str, device_args: list, cgse_cmd: str):
        """Initialisation of a stop command for a device Control Server.

        Args:
            device_id (str): Device identifier.
            device_args (list): Device arguments.
            cgse_cmd (str): CGSE command to start/stop the Control Server or to query its status.
        """

        super().__init__(device_id, device_args, cgse_cmd)

    @property
    def cmd(self) -> str:
        """Returns the full CGSE command to stop the Control Server.

        Returns: Full CGSE command to stop the Control Server.
        """

        cmd = f"{self._cgse_cmd} stop"

        if self.device_args:
            cmd = f"{cmd} {self.device_args[0]}"

        return cmd


class StatusCommand(ProcessCommand):
    """Command to query the status of a Control Server for a device."""

    def __init__(self, device_id: str, device_args: list = None, cgse_cmd: str = None):
        """Initialisation of a status command for a device Control Server.

        Args:
            device_id (str): Device identifier.
            device_args (list): Device arguments.
            cgse_cmd (str): CGSE command to start/stop the Control Server or to query its status.
        """

        super().__init__(device_id, device_args, cgse_cmd)

    @property
    def cmd(self) -> str:
        """Returns the full CGSE command to query the status of the Control Server.

        Returns: Full CGSE command to query the status of the Control Server.
        """

        if len(self.device_args) > 0:
            return f"{self._cgse_cmd} status {self.device_args}"
        else:
            return f"{self._cgse_cmd} status"


class ProcessManagerCommand(ClientServerCommand):
    """Client-server command for the Process Manager."""

    pass


class ProcessManagerInterface:
    def __init__(self):
        super().__init__()
        self.setup = load_setup()

    @dynamic_interface
    def get_core_processes(self) -> dict:
        """Returns a dictionary with the core CGSE processes.

        These processes should be running at all times, and can neither be started nor shut down from the Process
        Manager.  On an operational machine, these processes should be added to systemd to make sure they are
        re-started automatically if they are stopped.

        The keys in the dictionary are the names of the core processes (as they will be displayed in the PM UI).  The
        values are the names of the scripts as defined in the pyproject.toml file(s) under `[project.scripts]`.  Those
        can be used to start and stop the core processes, and to request their status.

        Returns: Dictionary with the core CGSE processes.
        """

        raise NotImplementedError

    # @dynamic_interface
    # def get_processing_processes(self):
    #
    #     raise NotImplementedError
    #
    # @dynamic_interface
    # def get_sut_processes(self):
    #
    #     raise NotImplementedError

    @dynamic_interface
    def get_devices(self) -> dict:
        """Returns a dictionary with the devices that are included in the setup.

        The device processes that are listed in the returned dictionary are the ones that are included in the setup
        that is currently loaded in the Configuration Manager.  The keys in the dictionary are taken from the
        "device_name" entries in the setup file.  The corresponding values consist of a tuple with the following
        entries:
            - `device` raw value (should be `Proxy` classes);
            - device identifier;
            - device arguments (optional);

        Returns: Dictionary with the devices that are included in the setup. The keys are the device name,
                 the values are tuples with the 'device' raw value, device identifier, and the (optional) device
                 arguments as a tuple.
        """

        raise NotImplementedError

    @dynamic_interface
    def get_device_ids(self) -> dict:
        """Returns a list with the identifiers of the devices that are included in the setup.

        The devices for which the identifiers are returned are the ones that are included in the setup that is currently
        loaded in the Configuration Manager.

        Returns: List with the identifiers of the devices that are included in the setup.
        """

        raise NotImplementedError

    @dynamic_interface
    def start_process(self, start_cmd: StartCommand) -> None:
        """Starts a process.

        Args:
            start_cmd (StartCommand): Command to start a process
        """

        raise NotImplementedError

    @dynamic_interface
    def stop_process(self, stop_cmd: StopCommand) -> None:
        """Stops a process.

        Args:
            stop_cmd (StopCommand): Command to stop a process
        """

        raise NotImplementedError

    @dynamic_interface
    def get_core_service_status(self, status_cmd: StatusCommand) -> dict:
        """Returns the status of a core service.

        Args:
            status_cmd (StatusCommand): Command to query the status of a core service

        Returns: Status of a core service
        """

        raise NotImplementedError

    @dynamic_interface
    def get_device_process_status(self, status_cmd: StatusCommand) -> dict:
        """Returns the status of a process.

        Args:
            status_cmd (StatusCommand): Command to query the status of a process

        Returns: Status of a process
        """

        raise NotImplementedError


class ProcessManagerController(ProcessManagerInterface):
    def __init__(self):
        super().__init__()

        # self._configuration = ConfigurationManagerProxy()

        if not is_storage_manager_active():
            LOGGER.error("No Storage Manager available!!!!")

    def get_core_processes(self) -> dict:
        core_processes = {}
        for ep in sorted(entry_points("cgse.process_management.core_services"), key=lambda x: x.name):
            core_processes[ep.name] = ep.value

        return core_processes

    def get_devices(self) -> dict:
        try:
            setup = load_setup()

            devices = {}
            devices = Setup.find_devices(setup, devices=devices)

            return devices

        except AttributeError:
            return {}

    def get_device_ids(self) -> dict:
        try:
            setup = load_setup()

            device_ids = {}
            device_ids = Setup.find_device_ids(setup, device_ids=device_ids)

            return device_ids

        except AttributeError:
            return {}

    def start_process(self, start_cmd: StartCommand):
        # subprocess.call(start_cmd.cmd, shell=True) -> PM hangs
        process = SubProcess("MyApp", [start_cmd.cmd], shell=True)
        process.execute()

    def stop_process(self, stop_cmd: StopCommand):
        subprocess.call(stop_cmd.cmd, shell=True)

    def get_core_service_status(self, status_cmd: StatusCommand) -> dict:
        output = subprocess.check_output(status_cmd.cmd, shell=True).decode("utf-8")
        cs_is_active = not ("inactive" in output or "not active" in output)

        return {"core_service_name": status_cmd.device_id, "core_service_is_active": cs_is_active}

    def get_device_process_status(self, status_cmd: StatusCommand) -> dict:
        output = subprocess.check_output(status_cmd.cmd, shell=True).decode("utf-8")
        # return output
        cs_is_active = not ("inactive" in output or "not active" in output)

        if cs_is_active:
            device_is_connected = not "not connected" in output
            is_simulator_mode = "simulator" in output

            return {
                "device_id": status_cmd.device_id,
                "cs_is_active": True,
                "device_is_connected": device_is_connected,
                "is_simulator_mode": is_simulator_mode,
            }
        else:
            return {"device_id": status_cmd.device_id, "cs_is_active": False}


class ProcessManagerProxy(Proxy, ProcessManagerInterface):
    """Proxy for process management, used to connect to the Process Manager Control Server and send commands remotely."""

    def __init__(
        self,
        protocol: str = PROTOCOL,
        hostname: str = HOSTNAME,
        port: int = COMMANDING_PORT,
        timeout: float = PROXY_TIMEOUT,
    ):
        """
        Initialisation of a new Proxy for Process Management.

        The connection details (transport protocol, hostname, and port) are by default taken from the
        settings file. When the `port` is 0 (zero) the endpoint is retrieved from the service registry.

        Args:
            protocol (str): Transport protocol
            hostname (str): Location of the control server (IP address)
            port (int): TCP port on which the Control Server is listening for commands
            timeout (float): number of fractional seconds before a timeout occurs
        """

        endpoint = get_endpoint(settings.SERVICE_TYPE, protocol, hostname, port)

        super().__init__(endpoint, timeout=timeout)



import pytest

from egse.listener import EVENT_ID
from egse.listener import Event
from egse.listener import Listeners


def test_listeners():
    listeners = Listeners()
    assert len(listeners) == 0

    # Success stories ----------------------------------------------------------

    listeners.add_listener({"name": "test_process_1"})
    assert len(listeners) == 1

    assert "test_process_1" in listeners.get_listener_names()

    listeners.remove_listener({"name": "test_process_1"})
    assert len(listeners) == 0

    assert "test_process_1" not in listeners.get_listener_names()

    listeners.add_listener({"name": "test_process_2"})
    listeners.add_listener({"name": "test_process_3"})

    assert "test_process_2" in listeners.get_listener_names()
    assert "test_process_3" in listeners.get_listener_names()

    # A KeyError will be raised because no Proxy was defined
    with pytest.raises(KeyError):
        listeners.notify_listeners(Event(EVENT_ID.ALL, "General notification"))

    # A KeyError will be raised because no Proxy was defined
    with pytest.raises(KeyError):
        listeners.notify_listeners(Event(event_id=EVENT_ID.SETUP, context={"new_item": "This is a new item."}))

    # Exceptions ---------------------------------------------------------------

    listeners.add_listener({"name": "test_process_1"})
    with pytest.raises(ValueError):
        listeners.add_listener({"name": "test_process_1"})

    with pytest.raises(ValueError):
        listeners.add_listener({"name": "test_process_2"})

    listeners.remove_listener({"name": "test_process_1"})
    with pytest.raises(ValueError):
        listeners.remove_listener({"name": "test_process_1"})

    with pytest.raises(ValueError):
        listeners.add_listener({})

    with pytest.raises(ValueError):
        listeners.add_listener({"name": "invalid_proxy", "proxy": "should be a Proxy class"})



import asyncio

from egse.env import bool_env
from egse.log import logger
from egse.settings import Settings
from egse.tempcontrol.keithley.daq6510_adev import DAQ6510

settings = Settings.load("Keithley DAQ6510")

VERBOSE_DEBUG = bool_env("VERBOSE_DEBUG")

SCPI_PORT = settings.get("PORT", 5025)
HOSTNAME = settings.get("HOSTNAME", "localhost")


async def test_a_scan():
    async with DAQ6510(HOSTNAME, SCPI_PORT) as daq:
        logger.info("Initialization: resetting device...")

        await daq.write("*RST")  # this also the user-defined buffer "test1"

        init_commands = [
            ('TRAC:MAKE "test1", 1000', False),  # create a new buffer
            # settings for channel 1 and 2 of slot 1
            ('SENS:FUNC "TEMP", (@101:102)', False),  # set the function to temperature
            ("SENS:TEMP:TRAN FRTD, (@101)", False),  # set the transducer to 4-wire RTD
            ("SENS:TEMP:RTD:FOUR PT100, (@101)", False),  # set the type of the 4-wire RTD
            ("SENS:TEMP:TRAN RTD, (@102)", False),  # set the transducer to 2-wire RTD
            ("SENS:TEMP:RTD:TWO PT100, (@102)", False),  # set the type of the 2-wire RTD
            ('ROUT:SCAN:BUFF "test1"', False),  # set the buffer for the scan
            ("ROUT:SCAN:CRE (@101:102)", False),  # create a scan list with channels 101 and 102
            ("ROUT:CHAN:OPEN (@101:102)", False),  # open the channels
            ("ROUT:STAT? (@101:102)", True),  # check if the channels are open
            ("ROUT:SCAN:STAR:STIM NONE", False),  # set the trigger to immediate
            ("ROUT:SCAN:ADD:SING (@101, 102)", False),  # add the channels to the scan list
            ("ROUT:SCAN:COUN:SCAN 1", False),  # set the number of scans to 1
            ("ROUT:SCAN:INT 1", False),  # set the scan interval to 1 second
        ]

        response = await daq.initialize(init_commands)
        if VERBOSE_DEBUG:
            logger.debug(f"Initialization response: {response}")

        # Read out the channels

        # daq.write('TRAC:CLE "test1"\n')

        for _ in range(10):
            await daq.write("INIT:IMM")
            await daq.write("*WAI")

            # Reading the data

            # When a trigger mode is running, these READ? commands can not be used.

            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")
            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")
            # time.sleep(1)
            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")
            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")

            # Read out the buffer

            response = await daq.trans('TRAC:DATA? 1, 2, "test1", CHAN, TST, READ')  # read 2 values from buffer "test1"
            if VERBOSE_DEBUG:
                print(f"{response = }")
            ch1, tst1, val1, ch2, tst2, val2 = response.decode().split(",")
            print(f"Channel: {ch1} Time: {tst1} Value: {float(val1):.4f}")
            print(f"Channel: {ch2} Time: {tst2} Value: {float(val2):.4f}")

            await asyncio.sleep(1.0)


if __name__ == "__main__":
    try:
        main_task = asyncio.run(test_a_scan())
    except KeyboardInterrupt:
        print("Caught KeyboardInterrupt, terminating.")



import sys
import time

from egse.log import logging
from egse.observation import ObservationContext
from egse.observation import building_block
from egse.observation import end_observation
from egse.observation import execute
from egse.observation import start_observation

logger = logging.getLogger("egse.test.script.observation_with_building_block")


def execute_start_observation_and_building_block():
    obsid = start_observation("Test observation with building block")
    logger.info(f"Started observation with obsid={obsid}")

    time.sleep(1.0)
    result = example_building_block()
    logger.info(f"Result of building block: {result}")
    time.sleep(1.0)

    end_observation()
    logger.info("Ended observation")


def execute_building_block():
    result = execute(example_building_block)
    logger.info(f"Result of building block: {result}")


def check_observation_context():
    ObservationContext().start_observation({"description": "Check observation context"})
    assert ObservationContext().get_level() == 0, "Observation level should be 0 at the start of the observation"
    ObservationContext().end_observation()


@building_block
def example_building_block() -> int:
    logger.info("Executing example building block")
    time.sleep(2.0)
    logger.info("Finishing example building block")
    return 42


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    execute_start_observation_and_building_block()
    execute_building_block()
    check_observation_context()



import datetime
import os

from egse.metrics import DataPoint
from egse.metrics import get_metrics_repo


def test_datapoint_as_dict_serializes_datetime_timestamp():
    timestamp = datetime.datetime(2026, 3, 25, 15, 12, 52, tzinfo=datetime.timezone.utc)

    point = DataPoint.measurement("camera_tm").field("temperature", 23.4).time(timestamp)

    payload = point.as_dict()

    assert payload["measurement"] == "camera_tm"
    assert payload["fields"]["temperature"] == 23.4
    assert payload["time"] == timestamp.timestamp()


def test_get_metrics_repo():
    token = os.environ.get("INFLUXDB3_AUTH_TOKEN")

    influxdb = get_metrics_repo("influxdb", {"host": "http://localhost:8181", "database": "ARIEL", "token": token})
    influxdb.connect()

    # Don't use a too large time interval here, or you will get an error like:
    # 'External error: Query would exceed file limit of 432 parquet files'
    result = influxdb.query(
        "SELECT * FROM cm WHERE time >= now() - INTERVAL '2 days' ORDER BY TIME DESC LIMIT 20", mode="pandas"
    )
    print(result)

    # result = influxdb.query("SHOW TABLES;")
    # result = influxdb.query("SHOW COLUMNS IN cm;", mode="pandas")
    # print(f"Columns in cm: {result}")

    result = influxdb.get_table_names()
    print(f"Tables in ARIEL: {result}")

    result = influxdb.get_column_names("cm")
    print(f"Columns in cm: {result}")

    influxdb.close()



import pickle
import types
from abc import ABC
from typing import Any

from egse.control import ControlServer
from egse.decorators import dynamic_interface
from egse.protocol import CommandProtocol
from egse.protocol import get_function
from egse.protocol import get_method
from egse.zmq_ser import bind_address


class _T:
    """A test class for the get_method() and get_function() functions."""

    _a = None

    def _m(self): ...

    @dynamic_interface
    def _di(self): ...


def test_get_method():
    assert get_method(_T, "_m") is None  # _m is a function for _T which is a class
    assert get_method(_T(), "") is None
    assert get_method(_T(), "None") is None
    assert get_method(_T(), None) is None

    assert get_method(_T(), "unknown") is None

    assert isinstance(get_method(_T(), "_m"), types.MethodType)
    assert isinstance(get_method(_T(), "_di"), types.MethodType)

    assert isinstance(get_method(_T(), "__getattribute__"), types.MethodWrapperType)


def test_get_function():
    assert get_function(_T, "") is None
    assert get_function(_T, "None") is None
    assert get_function(_T, None) is None

    assert get_function(_T, "unknown") is None
    assert get_function(_T, "_a") is None

    assert isinstance(get_function(_T, "_m"), types.FunctionType)
    assert isinstance(get_function(_T, "_di"), types.FunctionType)


class InternalCommunication:
    def __init__(self):
        self._status = 0
        self._pickle_string: bytes = bytes()

    def send(self, data):
        self._pickle_string = pickle.dumps(data)

    def receive(self) -> Any:
        """
        Returns an object based on the last command that was sent and based on
        the status of the internal communication.
        """
        data = pickle.loads(self._pickle_string)
        return data


class ControlServerMock(ControlServer):
    def __init__(self):
        super().__init__()
        self.mon_delay = 23

    def get_communication_protocol(self) -> str:
        return "tcp"

    def get_commanding_port(self) -> int:
        return 55555

    def get_service_port(self) -> int:
        return 55556

    def get_monitoring_port(self) -> int:
        return 55557


# Why are there only two abstract methods?
# - get_status()
# - get_bind_address()
# Why are the following methods also not abstract?
# - get_housekeeping()
# -
class _TestCommandProtocol(CommandProtocol):
    """A test class to unit test the CommandProtocol."""

    def __init__(self):
        super().__init__(ControlServerMock())
        self._comm = InternalCommunication()

    def get_status(self) -> dict:
        return super().get_status()

    def get_bind_address(self) -> str:
        return bind_address(self.control_server.get_communication_protocol(), self.control_server.get_commanding_port())

    def receive(self) -> Any:
        """Overwrites the parent class method"""
        response = self._comm.receive()
        return response

    def send(self, data):
        self._comm.send(data)


def test_command_protocol():
    tcp = _TestCommandProtocol()

    status = tcp.get_status()

    assert "timestamp" in status
    assert "PID" in status



"""
Test to verify the server is running in the background.

Start the test as follows (with pytest):

    $ uv run pytest -v --setup-show tests/test_server_running.py

Start the test wihout pytest:

    $ uv run py tests/test_server_running.py
"""

import asyncio
import json
import time

import pytest
import zmq
import zmq.asyncio
from fixtures.helpers import is_service_registry_running

from egse.log import logger
from egse.registry.backend import AsyncInMemoryBackend
from egse.registry.server import AsyncRegistryServer

TEST_REQ_PORT = 15556
TEST_PUB_PORT = 15557

SERVER_STARTUP_TIMEOUT = 5

pytestmark = pytest.mark.skipif(
    is_service_registry_running(), reason="This test starts its own registry server, so it can not be running."
)


async def server_health_check(zmq_context):
    # Verify server is ready by testing the health endpoint
    start_time = time.time()
    test_socket = zmq_context.socket(zmq.DEALER)
    test_socket.setsockopt(zmq.LINGER, 0)  # Don't wait for unsent messages on close()
    test_socket.setsockopt(zmq.IDENTITY, b"health-check-client")
    test_socket.connect(f"tcp://localhost:{TEST_REQ_PORT}")

    server_ready = False
    health_request = {"action": "health"}

    # Try several times to connect to the server
    for attempt in range(1, 11):  # 10 attempts
        logger.info(f"Attempt to connect to the server: {attempt}")
        try:
            if time.time() - start_time > SERVER_STARTUP_TIMEOUT:
                break

            # Send health check request
            await test_socket.send_multipart([b"REQ", json.dumps(health_request).encode()])

            # Wait for response
            if await test_socket.poll(timeout=1000):
                response_parts = await test_socket.recv_multipart()
                response_type = response_parts[0]
                logger.info(f"Received response type: {response_type=}")
                response_data = json.loads(response_parts[1].decode())
                logger.info(f"Received response data: {response_data=}")

                if response_data.get("success"):
                    server_ready = True
                    logger.info(f"Server ready after {attempt} attempts")
                    break

        except Exception as e:
            logger.error(f"ERROR: Attempt {attempt} failed: {e}")

        # Wait before trying again
        await asyncio.sleep(0.5)

    # Clean up test socket
    test_socket.close()

    return server_ready


async def start_server():
    """
    Session-scoped fixture for AsyncRegistryServer that verifies the server
    is ready before proceeding.
    """
    zmq_context = zmq.asyncio.Context()

    backend = AsyncInMemoryBackend()
    await backend.initialize()

    # Create server with specified ports
    server = AsyncRegistryServer(
        req_port=TEST_REQ_PORT,
        pub_port=TEST_PUB_PORT,
        backend=backend,
        cleanup_interval=1,  # Fast cleanup for testing
    )

    # Start the server in a task
    server_task = asyncio.create_task(server.start())

    server_ready = await server_health_check(zmq_context)

    if not server_ready:
        # Stop the server if it failed to start properly
        server.stop()
        try:
            await asyncio.wait_for(server_task, timeout=2.0)
        except asyncio.TimeoutError:
            pass

        raise RuntimeError(f"Server failed to start after {SERVER_STARTUP_TIMEOUT} seconds")

    logger.info("Server is ready...")

    # Yield the server for tests to use
    return server, server_task


async def stop_server(server, server_task):
    # Clean shutdown
    logger.info("Stopping server...")

    server.stop()

    try:
        await asyncio.wait_for(server_task, timeout=2.0)
    except asyncio.TimeoutError:
        logger.warning("WARNING: Server task didn't complete in time during shutdown")


@pytest.mark.asyncio
async def test_server_running():
    await main()


async def main():
    server, server_task = await start_server()

    await asyncio.sleep(10.0)

    await stop_server(server, server_task)


# This allows you to run the test without using pytest

if __name__ == "__main__":
    asyncio.run(main())



import sys
from pathlib import Path
from typing import Optional
from typing import Union

from PyQt5.QtCore import QPoint
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPainter
from PyQt5.QtSvg import QSvgRenderer
from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QCheckBox
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QWidget

from egse.resource import get_resource

BUTTON_DISABLED = 0
BUTTON_SELECTED = 1
BUTTON_NOT_SELECTED = 2
BUTTON_NO_CHANGE = 3
BUTTON_ACTIONED = 4


# Use the ToolTouchButton when you do not have a QToolBar to manage your buttons,
# but you have them in a QVBoxLayout or QHBoxLayout. The other TouchButton will not
# properly work if not used in a QToolBar, i.e. the icons will not have the proper size.


class ToolTouchButton(QToolButton):
    def __init__(
        self,
        width: int = 32,
        height: int = 32,
        name: str = None,
        status_tip: str = None,
        selected: Union[str, Path] = None,
        disabled: Union[str, Path] = None,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent=parent)

        self.setFixedSize(width, height)
        self.setCursor(Qt.PointingHandCursor)
        self.setText(name)
        self.setStatusTip(status_tip)

        self.status_tip = status_tip

        self.button_selected = selected
        self.button_disabled = disabled

        self.resource = {
            BUTTON_DISABLED: self.button_disabled,
            BUTTON_SELECTED: self.button_selected,
        }

        self.state = BUTTON_SELECTED
        self.disabled = False

        self.clicked.connect(self.handle_clicked)

    def handle_clicked(self):
        self.repaint()

    def print_clicked(self, *args, **kwargs):
        print(f"clicked: {args=}, {kwargs=}")
        print(f"         {self.text()=}")
        self.repaint()

    def print_pressed(self, *args, **kwargs):
        print(f"pressed: {args=}, {kwargs=}")

    def print_released(self, *args, **kwargs):
        print(f"released: {args=}, {kwargs=}")

    def setDisabled(self, flag: bool = True):
        self.disabled = flag
        super().setDisabled(flag)
        self.setStatusTip(f"{self.status_tip or ''} {'[DISABLED]' if flag else ''}")

    def disable(self):
        self.setDisabled(True)

    def enable(self):
        self.setDisabled(False)

    def hitButton(self, pos: QPoint):
        return self.contentsRect().contains(pos)

    def paintEvent(self, *args, **kwargs):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setPen(Qt.NoPen)

        self.drawCustomWidget(painter)

        painter.end()

    def drawCustomWidget(self, painter):
        renderer = QSvgRenderer()
        resource = self.resource[self.state if not self.disabled else BUTTON_DISABLED]
        renderer.load(str(resource))
        renderer.render(painter)


class TouchButton(QPushButton):
    def __init__(
        self,
        width: int = 32,
        height: int = 32,
        name: str = None,
        status_tip: str = None,
        selected: Union[str, Path] = None,
        disabled: Union[str, Path] = None,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent=parent)

        self.setFixedSize(width, height)
        self.setCursor(Qt.PointingHandCursor)
        self.setText(name)
        self.setStatusTip(status_tip)

        self.status_tip = status_tip

        self.button_selected = selected
        self.button_disabled = disabled

        self.resource = {
            BUTTON_DISABLED: self.button_disabled,
            BUTTON_SELECTED: self.button_selected,
        }

        self.state = BUTTON_SELECTED
        self.disabled = False

        self.clicked.connect(self.handle_clicked)

    def handle_clicked(self):
        self.repaint()

    def print_clicked(self, *args, **kwargs):
        print(f"clicked: {args=}, {kwargs=}")
        print(f"         {self.text()=}")
        self.repaint()

    def print_pressed(self, *args, **kwargs):
        print(f"pressed: {args=}, {kwargs=}")

    def print_released(self, *args, **kwargs):
        print(f"released: {args=}, {kwargs=}")

    def setDisabled(self, flag: bool = True):
        self.disabled = flag
        super().setDisabled(flag)
        self.setStatusTip(f"{self.status_tip or ''} {'[DISABLED]' if flag else ''}")

    def disable(self):
        self.setDisabled(True)

    def enable(self):
        self.setDisabled(False)

    def hitButton(self, pos: QPoint):
        return self.contentsRect().contains(pos)

    def paintEvent(self, *args, **kwargs):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setPen(Qt.NoPen)

        self.drawCustomWidget(painter)

        painter.end()

    def drawCustomWidget(self, painter):
        renderer = QSvgRenderer()
        resource = self.resource[self.state if not self.disabled else BUTTON_DISABLED]
        renderer.load(str(resource))
        renderer.render(painter)


class ToggleButton(QCheckBox):
    def __init__(
        self,
        width: int = 32,
        height: int = 32,
        name: str = None,
        status_tip: str = None,
        selected: Union[str, Path] = None,
        not_selected: Union[str, Path] = None,
        no_change: Union[str, Path] = None,
        disabled: Union[str, Path, list] = None,
        tristate: bool = False,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent=parent)

        self.setFixedSize(width, height)
        self.setCursor(Qt.PointingHandCursor)
        self.setTristate(tristate)
        self.setText(name)
        self.setStatusTip(status_tip)

        self.status_tip = status_tip

        self.max_states = 3 if tristate else 2

        self.button_selected = selected
        self.button_not_selected = not_selected
        self.no_change = no_change
        if isinstance(disabled, list):
            # this assumed we do not change the states definition anymore!
            self.button_disabled = [0, *disabled]
        else:
            self.button_disabled = [0, disabled, disabled, disabled]

        self.resource = {
            BUTTON_DISABLED: self.button_disabled,
            BUTTON_SELECTED: self.button_selected,
            BUTTON_NOT_SELECTED: self.button_not_selected,
            BUTTON_NO_CHANGE: self.no_change,
        }

        self.state = BUTTON_SELECTED
        self.disabled = False

        self.clicked.connect(self.handle_clicked)

    def __str__(self):
        return f"ToggleButton: name={self.text()}, tristate={self.isTristate()}, selected={self.is_selected()}"

    def handle_clicked(self):
        self.state = 1 if self.state == self.max_states else self.state + 1
        self.repaint()

    def setDisabled(self, flag: bool = True):
        self.disabled = flag
        super().setDisabled(flag)
        self.setStatusTip(f"{self.status_tip or ''} {'[DISABLED]' if flag else ''}")

    def disable(self):
        self.setDisabled(True)

    def enable(self):
        self.setDisabled(False)

    def is_selected(self):
        return self.state == BUTTON_SELECTED

    def set_selected(self, on: bool = True):
        self.state = BUTTON_SELECTED if on else BUTTON_NOT_SELECTED
        self.repaint()

    def hitButton(self, pos: QPoint):
        return self.contentsRect().contains(pos)

    def paintEvent(self, *args, **kwargs):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setPen(Qt.NoPen)

        self.drawCustomWidget(painter)
        painter.end()

    def drawCustomWidget(self, painter):
        renderer = QSvgRenderer()
        if not self.disabled:
            resource = self.resource[self.state]
        else:
            resource = self.resource[BUTTON_DISABLED][self.state]
        renderer.load(str(resource))
        renderer.render(painter)


if __name__ == "__main__":
    from PyQt5.QtWidgets import QMainWindow
    from PyQt5.QtWidgets import QFrame
    from PyQt5.QtWidgets import QVBoxLayout

    class MainWindow(QMainWindow):
        def __init__(self):
            super().__init__()
            self.resize(200, 200)

            self.container = QFrame()
            self.container.setObjectName("Container")
            self.layout = QVBoxLayout()

            # button_selected = get_resource(":/icons/cs-connected.svg")
            # button_not_selected = get_resource(":/icons/cs-not-connected.svg")
            # button_no_change = get_resource(":/icons/cs-connected-alert.svg")
            # button_disabled = get_resource(":/icons/cs-connected-disabled.svg")

            button_selected = get_resource(":/icons/start-button.svg")
            button_not_selected = get_resource(":/icons/stop-button.svg")
            button_no_change = get_resource(":/icons/plugged.svg")
            button_disabled = [
                get_resource(":/icons/start-button-disabled.svg"),
                get_resource(":/icons/stop-button-disabled.svg"),
                get_resource(":/icons/plugged-disabled.svg"),
            ]

            self.toggle = ToggleButton(
                name="CS-CONNECT",
                status_tip="connect-disconnect hexapod control server",
                selected=button_selected,
                not_selected=button_not_selected,
                no_change=button_no_change,
                disabled=button_disabled,
                tristate=True,
            )

            self.touch = TouchButton(
                name="HOME",
                status_tip="Perform a homing operation",
                selected=get_resource(":/icons/home.svg"),
                disabled=get_resource(":/icons/home-disabled.svg"),
            )

            self.layout.addWidget(self.toggle, Qt.AlignCenter, Qt.AlignCenter)
            self.layout.addWidget(self.touch, Qt.AlignCenter, Qt.AlignCenter)
            self.layout.addWidget(pb := QPushButton("disable"))
            self.container.setLayout(self.layout)
            self.setCentralWidget(self.container)

            self.pb = pb
            self.pb.setCheckable(True)
            self.pb.clicked.connect(self.toggle_disable)

            self.toggle.clicked.connect(self.print_clicked)
            self.touch.clicked.connect(self.print_clicked)

            # Use the following for further debugging

            # self.toggle.stateChanged.connect(self.print_state_changed)
            # self.toggle.pressed.connect(self.print_pressed)
            # self.toggle.released.connect(self.print_released)

            # self.touch.pressed.connect(self.print_pressed)
            # self.touch.released.connect(self.print_released)

            self.statusBar()

        def toggle_disable(self, checked: bool):
            self.toggle.disable() if checked else self.toggle.enable()
            self.touch.disable() if checked else self.touch.enable()
            self.pb.setText("enable" if checked else "disable")

        def print_state_changed(self, *args, **kwargs):
            sender = self.sender()
            print(f"stateChanged: {args=}, {kwargs=}")
            print(f"              {sender.isChecked()=} {sender.checkState()=}")

        def print_clicked(self, *args, **kwargs):
            sender = self.sender()
            print(f"clicked: {args=}, {kwargs=}")
            print(f"         {sender.state=}")
            print(f"         {sender.text()=}")

        def print_pressed(self, *args, **kwargs):
            print(f"pressed: {args=}, {kwargs=}")

        def print_released(self, *args, **kwargs):
            print(f"released: {args=}, {kwargs=}")

    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()

    sys.exit(app.exec_())



"""
Unit tests for AsyncRegistryServer and AsyncRegistryClient.

These tests verify the functionality of the ZeroMQ-based service registry components.

"""

import asyncio
import json
import time
import uuid

import pytest
import pytest_asyncio
import zmq
import zmq.asyncio
from fixtures.helpers import is_service_registry_running

from egse.log import logger
from egse.registry import MessageType
from egse.registry.backend import AsyncInMemoryBackend
from egse.registry.client import AsyncRegistryClient
from egse.registry.server import AsyncRegistryServer
from egse.system import type_name

# Constants for testing
TEST_REQ_PORT = 15556
TEST_PUB_PORT = 15557

# Wait timeout for the server to start (seconds)
SERVER_STARTUP_TIMEOUT = 5

pytestmark = pytest.mark.skipif(
    is_service_registry_running(), reason="This test starts the registry server, so it can not be running"
)

################################################################################
# Fixed Helper Functions
################################################################################


async def send_request(msg_type: MessageType, socket, request, timeout=5.0) -> dict:
    """
    Send a request to the server and get response with proper timeout.

    This improved version includes:
    - Timeout handling
    - Error handling
    - Diagnostic information
    """
    # Record start time for diagnostics
    start_time = time.time()

    try:
        # Send the request - add timeout to detect send failures
        try:
            await asyncio.wait_for(
                socket.send_multipart([msg_type.value, json.dumps(request).encode()]), timeout=timeout
            )
        except asyncio.TimeoutError:
            raise TimeoutError(f"Timeout sending request: {request}")

        logger.info(f"Request sent: {request}")

        # Wait for response with timeout
        if await socket.poll(timeout=timeout * 1000) == 0:
            raise TimeoutError(f"Timeout waiting for response to {request.get('action')} after {timeout} seconds")

        # Receive the response with timeout
        try:
            message_parts = await asyncio.wait_for(socket.recv_multipart(), timeout=timeout)

            if len(message_parts) >= 2:
                message_type = MessageType(message_parts[0])
                logger.debug(f"Received message type: {message_type}")
                message_data = message_parts[1]
                return json.loads(message_data)
            else:
                return {
                    "success": False,
                    "error": f"not enough message parts received: {len(message_parts)}",
                    "data": message_parts,
                }

        except asyncio.TimeoutError:
            raise TimeoutError(f"Timeout receiving response for {request.get('action')}")
        except json.JSONDecodeError as exc:
            raise ValueError(f"Invalid JSON response: {exc}")

    except KeyboardInterrupt:
        logger.info("Caught Keyboard Interrupt", exc_info=True)
        raise
    except Exception as exc:
        # Add diagnostics to the error
        elapsed = time.time() - start_time
        diagnostic_msg = f"Error in send_request after {elapsed:.2f}s: {str(exc)}"
        print(diagnostic_msg)  # Print for immediate feedback during test run
        raise RuntimeError(diagnostic_msg) from exc


async def wait_for_event(socket, timeout=2.0) -> dict | None:
    """
    Wait for an event from the server with timeout.

    This improved version includes better error handling.
    """
    # Wait for a message with timeout
    if await socket.poll(timeout=timeout * 1000) == 0:
        return None

    try:
        # Get the message with a timeout
        _, data = await asyncio.wait_for(socket.recv_multipart(), timeout=timeout)

        # Parse and return only the data part, type was for subscribers and is also included in data
        return json.loads(data.decode("utf-8"))

    except asyncio.TimeoutError:
        return None
    except Exception as e:
        print(f"Error in wait_for_event: {e}")
        return None


async def server_health_check(zmq_context) -> bool:
    """server_health_check
    Verify if the server is ready by testing the health endpoint/action.
    """

    start_time = time.time()
    test_socket = zmq_context.socket(zmq.DEALER)
    test_socket.setsockopt(zmq.LINGER, 0)  # Don't wait for unsent messages on close()
    test_socket.setsockopt(zmq.IDENTITY, "health-check-id".encode())
    test_socket.connect(f"tcp://localhost:{TEST_REQ_PORT}")

    server_ready = False
    msg_type = MessageType.REQUEST_WITH_REPLY
    health_request = {"action": "health"}

    # Try several times to connect to the server
    for attempt in range(1, 11):  # 10 attempts
        logger.info(f"Attempt to connect to the server: {attempt}")
        try:
            if time.time() - start_time > SERVER_STARTUP_TIMEOUT:
                break

            # Send health check request
            await test_socket.send_multipart([msg_type.value, json.dumps(health_request).encode()])

            # Wait for response
            if await test_socket.poll(timeout=1000):
                message_parts = await test_socket.recv_multipart()

                if len(message_parts) >= 2:
                    message_type = MessageType(message_parts[0])
                    message_data = message_parts[1]

                    response = json.loads(message_data)

                    if response.get("success"):
                        server_ready = True
                        logger.info(f"Server ready after {attempt} attempts")
                        break

        except Exception as exc:
            logger.error(f"ERROR: Attempt {attempt} failed: {exc}", exc_info=True)

        # Wait before trying again
        await asyncio.sleep(0.5)

    # Clean up test socket
    test_socket.close()

    return server_ready


################################################################################
# Fixtures
################################################################################


@pytest_asyncio.fixture
async def in_memory_backend():
    """Function-scoped fixture for an in-memory backend."""
    backend = AsyncInMemoryBackend()
    await backend.initialize()
    yield backend
    await backend.close()


@pytest_asyncio.fixture
async def server(in_memory_backend, zmq_context):
    """
    Function-scoped fixture for AsyncRegistryServer that verifies the server
    is ready before proceeding.
    """
    # Create server with specified ports
    server = AsyncRegistryServer(
        req_port=TEST_REQ_PORT,
        pub_port=TEST_PUB_PORT,
        backend=in_memory_backend,
        cleanup_interval=1,  # Fast cleanup for testing
    )

    # Start the server in a task to run in the background
    server_task = asyncio.create_task(server.start())

    server_ready = await server_health_check(zmq_context)

    if not server_ready:
        # Stop the server if it failed to start properly and raise a RuntimeError
        server.stop()
        try:
            await asyncio.wait_for(server_task, timeout=2.0)
        except asyncio.TimeoutError:
            pass
        except Exception as exc:
            msg = f"Caught {type_name(exc)}: {exc}"
            if str(exc).startswith("Address already in use"):
                msg += " – the Registry server is probably already running."
            logger.error(msg)

        raise RuntimeError(f"Server failed to start after {SERVER_STARTUP_TIMEOUT} seconds")

    logger.info("Server is ready...")

    yield server

    # Clean shutdown
    logger.info("Stopping server...")
    server.stop()
    try:
        await asyncio.wait_for(server_task, timeout=2.0)
    except asyncio.TimeoutError:
        logger.warning("WARNING: Server task didn't complete in time during shutdown")


@pytest_asyncio.fixture
async def client(server, zmq_context):
    """
    Function-scoped fixture for AsyncRegistryClient with proper setup and teardown.
    """
    with AsyncRegistryClient(
        registry_req_endpoint=f"tcp://localhost:{TEST_REQ_PORT}",
        registry_sub_endpoint=f"tcp://localhost:{TEST_PUB_PORT}",
        timeout=5000,  # 5 second timeout
    ) as client:
        # Verify client can connect to server
        health = await client.health_check()
        if not health:
            raise RuntimeError("Client failed to connect to server")

        yield client

        # Proper cleanup
        # await client.close()


@pytest_asyncio.fixture
async def zmq_context():
    """
    Function-scoped fixture that provides the ZeroMQ Context.
    """
    context = zmq.asyncio.Context().instance()
    logger.info("Yielding zmq context...")
    yield context
    logger.info("Terminating zmq context...")
    context.term()


@pytest_asyncio.fixture
async def req_socket(zmq_context):
    """
    Function-scoped fixture for a DEALER socket with proper timeout settings.
    """
    socket = zmq_context.socket(zmq.DEALER)
    # Set timeouts to avoid hanging
    socket.setsockopt(zmq.LINGER, 0)  # Don't wait when closing
    socket.setsockopt(zmq.RCVTIMEO, 5000)  # 5 second receive timeout
    socket.setsockopt(zmq.SNDTIMEO, 5000)  # 5 second send timeout
    socket.setsockopt(zmq.IDENTITY, f"client-{uuid.uuid4()}".encode())
    socket.connect(f"tcp://localhost:{TEST_REQ_PORT}")

    yield socket

    # Properly close the socket
    socket.close(linger=0)


@pytest_asyncio.fixture
async def sub_socket(zmq_context):
    """
    Function-scoped fixture for a SUB socket with proper timeout settings.
    """
    socket = zmq_context.socket(zmq.SUB)
    # Set timeouts
    socket.setsockopt(zmq.LINGER, 0)
    socket.setsockopt(zmq.RCVTIMEO, 5000)

    socket.connect(f"tcp://localhost:{TEST_PUB_PORT}")
    socket.setsockopt_string(zmq.SUBSCRIBE, "")  # Subscribe to all messages

    # Wait a moment for subscription to be established
    await asyncio.sleep(0.1)

    yield socket

    # Properly close the socket
    socket.close(linger=0)


@pytest.mark.asyncio
async def test_server_initialization(server):
    """Test that the server initializes correctly."""
    assert server.req_port == TEST_REQ_PORT
    assert server.pub_port == TEST_PUB_PORT
    assert server._running is True


@pytest.mark.asyncio
async def test_server_is_running(server):
    # Sleep for 10s and investigate the log, server should have printed some log messages
    logger.info("Sleeping for 10s, server shall post some debug log messages...")
    await asyncio.sleep(10.0)


@pytest.mark.asyncio
async def test_server_health_check(server, req_socket):
    """Test the server's health check."""
    response = await send_request(MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "health"})

    assert response["success"] is True
    assert response["status"] == "ok"
    assert "timestamp" in response


@pytest.mark.asyncio
async def test_server_handles_invalid_request(server, req_socket):
    """Test the server's handling of invalid requests."""
    # Missing action
    response = await send_request(MessageType.REQUEST_WITH_REPLY, req_socket, {})
    assert response["success"] is False
    assert "Missing required field: action" in response["error"]

    # Unknown action
    response = await send_request(MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "unknown_action"})
    assert response["success"] is False
    assert "Unknown action" in response["error"]

    # Invalid JSON - can't test directly with send_request as it uses json.dumps
    # This would require a lower-level test with raw socket


@pytest.mark.asyncio
async def test_server_register_service(server, req_socket, sub_socket):
    """Test registering a service."""

    service_info = {
        "name": "test-service",
        "host": "localhost",
        "port": 8080,
        "type": "test",
        "tags": ["test"],
        "metadata": {"version": "1.0.0"},
    }

    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "register", "service_info": service_info}
    )

    assert response["success"] is True
    assert "service_id" in response

    # Verify a registration event was published
    event = await wait_for_event(sub_socket)
    assert event is not None
    assert event["type"] == "register"
    assert "service_id" in event["data"]
    assert event["data"]["service_info"]["name"] == "test-service"


@pytest.mark.asyncio
async def test_server_get_service(server, req_socket):
    """Test getting a service by ID."""

    service_info = {"name": "test-service", "host": "localhost", "port": 8080}

    register_response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "register", "service_info": service_info}
    )

    service_id = register_response["service_id"]

    # Now get the service
    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "get", "service_id": service_id}
    )

    assert response["success"] is True
    assert response["service"]["id"] == service_id
    assert response["service"]["name"] == "test-service"
    assert response["service"]["host"] == "localhost"
    assert response["service"]["port"] == 8080

    # Try to get a non-existent service
    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "get", "service_id": "nonexistent"}
    )

    assert response["success"] is False
    assert "Service not found" in response["error"]


@pytest.mark.asyncio
async def test_server_list_services(server, req_socket):
    """Test listing services."""

    # Register a couple of services
    await send_request(
        MessageType.REQUEST_WITH_REPLY,
        req_socket,
        {
            "action": "register",
            "service_info": {"name": "service-1", "host": "localhost", "port": 8081, "type": "type-a"},
        },
    )

    await send_request(
        MessageType.REQUEST_WITH_REPLY,
        req_socket,
        {
            "action": "register",
            "service_info": {"name": "service-2", "host": "localhost", "port": 8082, "type": "type-b"},
        },
    )

    # List all services
    response = await send_request(MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "list"})

    assert response["success"] is True
    assert len(response["services"]) >= 2

    # List services by type
    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "list", "service_type": "type-a"}
    )

    assert response["success"] is True
    assert len(response["services"]) >= 1
    assert any(s["name"] == "service-1" for s in response["services"])
    assert not any(s["name"] == "service-2" for s in response["services"])


@pytest.mark.asyncio
async def test_server_discover_service(server, req_socket):
    """Test discovering a service by type."""

    await send_request(
        MessageType.REQUEST_WITH_REPLY,
        req_socket,
        {
            "action": "register",
            "service_info": {"name": "test-service", "host": "localhost", "port": 8080, "type": "discovery-test"},
        },
    )

    # Discover a service
    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "discover", "service_type": "discovery-test"}
    )

    assert response["success"] is True
    assert response["service"]["name"] == "test-service"
    assert response["service"]["host"] == "localhost"
    assert response["service"]["port"] == 8080

    # Try to discover a non-existent service type
    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "discover", "service_type": "nonexistent"}
    )

    assert response["success"] is False
    assert "No healthy service" in response["error"]


@pytest.mark.asyncio
async def test_server_deregister_service(server, req_socket, sub_socket):
    """Test de-registering a service."""

    register_response = await send_request(
        MessageType.REQUEST_WITH_REPLY,
        req_socket,
        {"action": "register", "service_info": {"name": "test-service", "host": "localhost", "port": 8080}},
    )

    service_id = register_response["service_id"]

    # Now deregister the service
    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "deregister", "service_id": service_id}
    )

    assert response["success"] is True

    # Verify a de-registration event was published, the first event is the
    # registration which we will skip.
    event = await wait_for_event(sub_socket)
    assert event is not None
    assert event["type"] == "register"
    event = await wait_for_event(sub_socket)
    assert event is not None
    assert event["type"] == "deregister"
    assert event["data"]["service_id"] == service_id

    # Verify the service was deregistered
    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "get", "service_id": service_id}
    )

    assert response["success"] is False


@pytest.mark.asyncio
async def test_server_renew_service(server, req_socket):
    """Test renewing a service's TTL."""

    register_response = await send_request(
        MessageType.REQUEST_WITH_REPLY,
        req_socket,
        {"action": "register", "service_info": {"name": "test-service", "host": "localhost", "port": 8080}, "ttl": 5},
    )

    service_id = register_response["service_id"]

    # Get the initial service data
    service_response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "get", "service_id": service_id}
    )

    initial_heartbeat = service_response["service"]["last_heartbeat"]

    # Wait a moment
    await asyncio.sleep(1.0)

    # Renew the service
    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "renew", "service_id": service_id}
    )

    assert response["success"] is True

    # Verify the heartbeat was updated
    service_response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "get", "service_id": service_id}
    )

    assert service_response["service"]["last_heartbeat"] > initial_heartbeat


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_server_cleanup_expired_services(req_socket, sub_socket, server):
    """Test cleanup of expired services."""

    # Register a service with a short TTL
    register_response = await send_request(
        MessageType.REQUEST_WITH_REPLY,
        req_socket,
        {
            "action": "register",
            "service_info": {"name": "short-lived-service", "host": "localhost", "port": 8080},
            "ttl": 2,  # Very short TTL
        },
    )

    service_id = register_response["service_id"]

    # Wait for the service to expire and be cleaned up
    # Should take about 2-3 seconds (TTL + cleanup interval)
    expire_event = None
    for _ in range(5):  # Try a few times
        event = await wait_for_event(sub_socket, timeout=1.0)
        if event and event["type"] == "expire" and event["data"]["service_id"] == service_id:
            expire_event = event
            break

    assert expire_event is not None, "Service expiration event not received"

    # Verify the service was removed
    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "get", "service_id": service_id}
    )

    assert response["success"] is False, "Service should have been removed"


@pytest.mark.asyncio
async def test_server_shutdown(server):
    """Test server shutdown."""

    assert server._running is True

    # Stop the server
    server.stop()

    await asyncio.sleep(0.5)  # Allow the server some time to shut down

    # Verify the server is no longer running
    assert server._running is False


# ######################################################################################################################
# Tests for AsyncRegistryClient
# ######################################################################################################################


@pytest.mark.asyncio
async def test_async_client_initialization(client):
    """Test that the client initializes correctly."""
    assert client.registry_req_endpoint == f"tcp://localhost:{TEST_REQ_PORT}"
    assert client.registry_sub_endpoint == f"tcp://localhost:{TEST_PUB_PORT}"


@pytest.mark.asyncio
async def test_async_client_register(client, server):
    """Test service registration."""

    service_id = await client.register(
        "test-client-service", "localhost", 8080, service_type="test-client", metadata={"version": "1.0.0"}
    )

    assert service_id is not None
    assert client._service_id == service_id
    assert client._service_info is not None
    assert client._service_info["name"] == "test-client-service"
    assert client._service_info["type"] == "test-client"


@pytest.mark.asyncio
async def test_client_deregister(client, server):
    """Test service de-registration."""

    # First register
    service_id = await client.register("test-client-service", "localhost", 8080)

    assert service_id is not None

    # Then deregister
    success = await client.deregister()

    assert success is True
    assert client._service_id is None


@pytest.mark.asyncio
async def test_client_heartbeat(client, server, req_socket):
    """Test service heartbeat."""

    # Register the service
    service_id = await client.register("heartbeat-test-service", "localhost", 8080, ttl=5)

    # Get initial heartbeat time
    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "get", "service_id": service_id}
    )

    initial_heartbeat = response["service"]["last_heartbeat"]

    # Start heartbeat with a short interval
    await client.start_heartbeat(interval=1)

    # Wait for heartbeat to happen
    await asyncio.sleep(1.5)

    # Check if heartbeat was updated
    response = await send_request(
        MessageType.REQUEST_WITH_REPLY, req_socket, {"action": "get", "service_id": service_id}
    )

    assert response["service"]["last_heartbeat"] > initial_heartbeat

    # Stop heartbeat
    await client.stop_heartbeat()


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_client_event_listener(client, server, req_socket):
    """Test the client's event listener."""

    # Create a mocked event handler
    event_received = asyncio.Event()
    event_data = {}

    async def event_handler(data):
        event_data.update(data)
        event_received.set()

    # Register the event handler and start the listener
    client.on_event("register", event_handler)
    await client.start_event_listener()

    # Register a service to trigger an event
    await send_request(
        MessageType.REQUEST_WITH_REPLY,
        req_socket,
        {"action": "register", "service_info": {"name": "event-test-service", "host": "localhost", "port": 8080}},
    )

    # Wait for the event to be processed
    await asyncio.wait_for(event_received.wait(), timeout=5.0)

    # Verify the event was received
    assert "service_id" in event_data
    assert "service_info" in event_data
    assert event_data["service_info"]["name"] == "event-test-service"

    # Stop the event listener
    await client.stop_event_listener()


@pytest.mark.asyncio
async def test_client_discover_service(client, server):
    """Test service discovery."""

    # Register a service through the server
    service_id = await client.register("discovery-test-service", "localhost", 8080, service_type="discovery-test")

    # Discover a service
    service = await client.discover_service("discovery-test")

    assert service is not None
    assert service["id"] == service_id
    assert service["name"] == "discovery-test-service"
    assert service["host"] == "localhost"
    assert service["port"] == 8080

    # Try to discover a non-existent service type
    service = await client.discover_service("nonexistent")
    assert service is None


@pytest.mark.asyncio
async def test_client_get_service(client, server):
    """Test getting a service by ID."""

    service_id = await client.register("get-test-service", "localhost", 8080)

    # Get the service
    service = await client.get_service(service_id)

    assert service is not None
    assert service["id"] == service_id
    assert service["name"] == "get-test-service"

    # Try to get a non-existent service
    service = await client.get_service("nonexistent")
    assert service is None


@pytest.mark.asyncio
async def test_client_list_services(client, server):
    """Test listing services."""

    # Register some services
    await client.register("list-test-service-1", "localhost", 8081, service_type="type-a")

    await client.register("list-test-service-2", "localhost", 8082, service_type="type-b")

    # List all services
    services = await client.list_services()

    assert len(services) >= 2
    assert any(s["name"] == "list-test-service-1" for s in services)
    assert any(s["name"] == "list-test-service-2" for s in services)

    # List services by type
    services = await client.list_services("type-a")

    assert len(services) >= 1
    assert any(s["name"] == "list-test-service-1" for s in services)
    assert not any(s["name"] == "list-test-service-2" for s in services)


@pytest.mark.asyncio
async def test_client_health_check(client, server):
    """Test health check."""
    health = await client.health_check()
    assert health is True
